import numpy as np
import fitsio
import roboscheduler.cadence
import sys

try:
    import sdssdb.peewee.sdss5db.opsdb as opsdb
    import sdssdb.peewee.sdss5db.targetdb as targetdb
    _database = True
except:
    _database = False


# Class to define a singleton
class FieldsSingleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(FieldsSingleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Fields(object, metaclass=FieldsSingleton):
    """List of fields. Initialized empty; attributes will be set
    via fromfits() or fromdb().

    Parameters:
    ----------

    plan : string
        robostrategy version, used for db access

    observatory : string
        APO or LCO, used for db access

    Attributes:
    ----------

    nfields : np.int32
        number of fields

    racen : ndarray of np.float64
        right ascension of each design (J2000 deg)

    deccen : ndarray of np.float64
        declination of each design (J2000 deg)

    field_id : ndarray of np.int32
        id for each field

    cadence : list of string
        cadence for each field

    observations : list of ndarray of np.int32
        for each field, indices of observations


    slots : ndarray of np.int32
        Nx2x24, corresponding to robostrategy planned LST bright/dark plan

    lstObserved : ndarray of np.int32
        Nx24 observations in each lst per field;
        respecting robostrategy plan leads to an efficient survey

    cadencelist : robostrategy.cadence.CadenceList singleton
        the cadenceList singleton, for easy access

    """
    def __init__(self, plan=None, observatory=None):
        self.plan = plan
        self.observatory = observatory
        self.nfields = 0
        self.racen = np.zeros(0, dtype=np.float64)
        self.deccen = np.zeros(0, dtype=np.float64)
        self.nfilled = np.zeros(0, dtype=np.float64)
        self.field_id = np.zeros(0, dtype=np.int32)
        # self.nextmjd = np.zeros(0, dtype=np.float64)
        self.epoch_idx = np.zeros(0, dtype=np.int32)
        self.notDone = np.ones(0, dtype=np.bool)  # true is not done to skip the invert elsewhere
        self.cadence = []
        self.observations = []
        self.slots = []
        self.flag = []
        self.lstObserved = np.zeros(0, dtype=np.int32)
        self.cadencelist = roboscheduler.cadence.CadenceList()
        self._validCadance = None
        self._obsPlan = None
        self._lstPlan = None
        self._lunationPlan = None
        self._hist = None
        self._database = _database
        return

    def setPriorities(self):
        scale = [10 if "bhm_rm" in c else 1 for c in self.cadence]
        self.basePriority = self.basePriority * np.array(scale)
        return

    def fromarray(self, fields_array=None):
        self.nfields = len(fields_array)
        self.racen = fields_array['racen']
        self.deccen = fields_array['deccen']
        self.nfilled = fields_array['nfilled']
        self.field_id = fields_array['field_id']
        self.cadence = [c.strip().decode() for c in fields_array['cadence']]
        self.slots = fields_array['slots_exposures']
        self.lstObserved = np.zeros((len(self.slots), 24), dtype=np.int32)
        self.observations = [np.zeros(0, dtype=np.int32)] * self.nfields
        self.icadence = np.zeros(self.nfields, dtype=np.int32)
        # self.nextmjd = np.zeros(self.nfields, dtype=np.float64)
        self.epoch_idx = np.zeros(self.nfields, dtype=np.float64)
        self.basePriority = np.ones(self.nfields) * 200
        self.notDone = np.ones(self.nfields, dtype=np.bool)
        if "flag" in fields_array.dtype.names:
            self.flag = fields_array["flag"]
        else:
            self.flag = np.zeros(self.nfields)
        self.setPriorities()
        return

    def fromfits(self, filename=None):
        """Load a fits file into this Fields object,
           likely an RS-Assignments file.

        Parameters:
        ----------

        filename : string
            the full path to the fits file to be loaded
        """
        self._database = False
        fields_model = [('field_id', np.int32),
                        ('racen', np.float64),
                        ('deccen', np.float64),
                        ('nfilled', np.int32),
                        ('flag', np.int32),
                        ('slots_exposures', np.int32, (24, 2)),
                        ('cadence', np.dtype('a20'))]

        fits_dat = fitsio.read(filename)

        self.fields_fits = np.zeros(len(fits_dat["fieldid"]), dtype=fields_model)

        self.fields_fits["field_id"] = fits_dat["fieldid"]
        self.fields_fits["racen"] = fits_dat["racen"]
        self.fields_fits["deccen"] = fits_dat["deccen"]
        self.fields_fits["nfilled"] = fits_dat["nfilled"]
        self.fields_fits["slots_exposures"] = fits_dat["slots_exposures"]
        self.fields_fits["cadence"] = fits_dat["cadence"]

        self.fromarray(self.fields_fits)
        return

    # def add_observations(self, mjd=None, fieldidx=None, iobs=None,
    #                      lst=None, epoch_idx=None):
    #     # """Add an observation, used in simulations.

    #     # Parameters:
    #     # ----------

    #     # mjd : float
    #     #     MJD of the observation
    #     # fieldidx : integer
    #     #     index into Fields of the field observed
    #     # iobs : integer
    #     #     index into Scheduler.observations
    #     # lst : float
    #     #     lst at time of observation
    #     # epoch_idx : integer
    #     #     which epoch was being observed
    #     # """
    #     # self._hist[self.field_id[fieldidx]].append(mjd)

    #     self.observations[fieldidx] = np.append(self.observations[fieldidx], iobs)
    #     self.icadence[fieldidx] = epoch_idx
    #     cadence = self.cadencelist.cadences[self.cadence[fieldidx]]
    #     if(self.icadence[fieldidx] < cadence.nepochs):
    #         self.nextmjd[fieldidx] = (mjd +
    #                                   cadence.delta_min[self.icadence[fieldidx]])
    #     else:
    #         self.nextmjd[fieldidx] = 100000.
    #         self.icadence[fieldidx] = cadence.nepochs - 1
    #     int_lst = int(np.round(lst/15, 0))
    #     if int_lst == 24:
    #         int_lst = 0
    #     self.lstObserved[fieldidx][int_lst] += 1
    #     return

    @property
    def validCadence(self):
        if self._validCadance is None:
            assert len(self.cadence) > 0, "no field cadences!"
            assert len(self.cadencelist.cadences) > 0, "no cadences in cadencelist!"
            self._validCadance = np.array([c in self.cadencelist.cadences
                                           for c in self.cadence])
        return self._validCadance

    @property
    def lstPlan(self):
        if self._lstPlan is None:
            self._lstPlan = np.array([np.sum(s, axis=1) for s in self.slots])
        return self._lstPlan

    @property
    def hist(self):
        # """The observation history of the field. Unlike Fields.observations,
        # this property will be used on sky.

        # Attributes:
        # ----------

        # _hist : dict of lists
        #     A dictionary with field_id keys containing lists of obs mjds for
        #     each field.
        # """
        if self._hist is None:
            self._hist = {f: list() for f in self.field_id}
            if self._database:
                versionDB = targetdb.Version()
                ver = versionDB.get(plan=self.plan)

                obsDB = targetdb.Observatory()
                obs = obsDB.get(label=self.observatory.upper())

                Field = targetdb.Field
                Design = targetdb.Design
                Status = opsdb.CompletionStatus
                d2s = opsdb.DesignToStatus
                done = Status.get(label="done")

                dbfields = Field.select(d2s.mjd, Field.field_id)\
                                .join(Design)\
                                .join(d2s, on=(Design.pk == d2s.design_pk))\
                                .where((Field.version == ver) &
                                       (Field.observatory == obs),
                                       (d2s.status == done)).dicts()

                for d in dbfields:
                    self._hist[d["field_id"]].append(d["mjd"])

            for i in range(len(self.field_id)):
                self.checkCompletion(i)

        return self._hist

    def checkCompletion(self, fieldidx):
        """evaluate field completion
        """
        if not self.notDone[fieldidx]:
            # already done
            return
        field_id = self.field_id[fieldidx]
        cadence = self.cadencelist.cadences[self.cadence[fieldidx]]
        tol = cadence.max_length[int(self.epoch_idx[fieldidx])]
        obs_epochs, begin_last_epoch = epochs_completed(self.hist[field_id],
                                                        tolerance=tol)
        self.epoch_idx[fieldidx] = int(obs_epochs)
        if obs_epochs >= cadence.nepochs:
            self.notDone[fieldidx] = False

    def completeDesign(self, field_idx, mjd, lst, iobs):
        """Maybe just for sims but need a way to add mjds to _hist[field_id]
        """
        self.observations[field_idx] = np.append(self.observations[field_idx], iobs)

        field_id = self.field_id[field_idx]

        self._hist[field_id].append(mjd)

        int_lst = int(np.round(lst/15, 0))
        if int_lst == 24:
            int_lst = 0
        self.lstObserved[field_idx][int_lst] += 1

        self.checkCompletion(field_idx)

    def fromdb(self, version=None):
        """Load this Fields object with fields from the targetdb

        Parameters:
        ----------

        version : string
            db version to grab, if Fields.plan is not set. If passed
            Fields.plan will be reset to version
        """
        if(self._database is False):
            print("No database available.")
            return()

        if version is None:
            version = self.plan
        else:
            self.plan = version
        assert version is not None, "must specify version!"

        fields_model = [('field_id', np.int32),
                        ('racen', np.float64),
                        ('deccen', np.float64),
                        ('nfilled', np.int32),
                        ('flag', np.int32),
                        ('slots_exposures', np.int32, (24, 2)),
                        ('cadence', np.dtype('a20'))]

        versionDB = targetdb.Version()
        ver = versionDB.get(plan=version)

        obsDB = targetdb.Observatory()
        obs = obsDB.get(label=self.observatory.upper())

        Field = targetdb.Field
        dbfields = Field.select().where(Field.version == ver,
                                        Field.observatory == obs)

        fieldid = list()
        racen = list()
        deccen = list()
        slots_exposures = list()
        cadence = list()
        flags = list()

        for field in dbfields:
            fieldid.append(field.field_id)
            racen.append(field.racen)
            deccen.append(field.deccen)
            slots_exposures.append(field.slots_exposures)
            cadence.append(field.cadence.label)
            if len(field.priority) > 0:
                if field.priority[0].label == "top":
                    flags.append(1)
                elif field.priority[0].label == "disabled":
                    flags.append(-1)
            else:
                flags.append(0)

        fields = np.zeros(len(dbfields), dtype=fields_model)

        fields["field_id"] = fieldid
        fields["racen"] = racen
        fields["deccen"] = deccen
        fields["flag"] = flags
        fields["slots_exposures"] = slots_exposures
        fields["cadence"] = cadence

        self.fromarray(fields_array=fields)

        self.cadencelist.fromdb()

    def lstWeight(self, lst, field_idx=None):
        # field id corresponds to indx, so fields is id/indx
        # as is everywhere, but just to keep reminding me...
        assert lst > 0 and lst < 24, "lst must be in hours!"
        diffs = []
        if field_idx is not None:
            lst_obs = self.lstObserved[field_idx]
            lst_plan = self.lstPlan[field_idx]
        else:
            lst_obs = self.lstObserved
            lst_plan = self.lstPlan

        for o, p in zip(lst_obs, lst_plan):
            done = p - o
            elligible = np.where(done > 0)[0]
            if len(elligible) == 0:
                diffs.append(12.)
                continue

            diff = lstDiff(elligible, lst*np.ones(len(elligible)))
            # at the moment we don't care which it lines up with
            diffs.append(np.min(diff))

        return np.array(diffs)

    def toarray(self):
        """Return cadences as a record array

        Returns:
        -------

        fields : ndarray
            information on each field
        """
        maxn = np.array([len(x) for x in self.observations]).max()
        if(maxn == 1):
            maxn = 2
        fields0 = [('fieldid', np.int32),
                   ('racen', np.float64),
                   ('deccen', np.float64),
                   ('cadence', np.dtype('a20')),
                   ('nobservations', np.int32),
                   ('observations', np.int32, maxn)]
        fields = np.zeros(self.nfields, dtype=fields0)
        fields['fieldid'] = self.field_id
        fields['racen'] = self.racen
        fields['deccen'] = self.deccen
        for indx in np.arange(self.nfields):
            fields['cadence'][indx] = self.cadence[indx]
            fields['nobservations'][indx] = len(self.observations[indx])
            if(fields['nobservations'][indx] > 0):
                fields['observations'][indx][0:fields['nobservations'][indx]] = self.observations[indx]
        return(fields)

    def getidx(self, fieldid):
        # return idx into array of fieldid
        return int(np.where(self.field_id == fieldid)[0])


def lstDiffSingle(a, b):
    """Intelligently find difference in 2 lsts

    Parameters:
    -----------

    a : np.float32, float
        first lst in hours
    b : np.float32, float
        second lst in hours

    Returns:
    --------

    diff: float
        the absolute difference

    Comments:
    ---------

    """

    if a < b:
        return min(b - a, (a + 24) - b)

    else:
        # a must be bigger
        return min(a - b, (b + 24) - a)


def lstDiff(a, b):
    """wrap lst math to handle arrays

    Parameters:
    -----------

    a : ndarray or list of float32
        first lst in hours
    b : ndarray or list of float32
        second lst in hours

    Returns:
    --------

    diff: ndarray of float32
        the absolute differences

    Comments:
    ---------

    """

    assert len(a) == len(b), "can't compare arrays of different size!"

    return np.array([lstDiffSingle(i, j) for i, j in zip(a, b)])


def epochs_completed(mjd_past, tolerance=0.5):
    """Calculate # of observed epochs, allowing
    for more exposures than planned.

    tolerance HAS CHANGED TO DAYS
    """
    if len(mjd_past) == 0:
        return 0, 0

    epoch_idx = 0

    tolerance = tolerance
    begin_last_epoch = mjd_past[0]

    obs_epochs = 1
    prev = begin_last_epoch
    for m in mjd_past:
        delta = m - prev
        if delta < tolerance:
            continue
        else:
            obs_epochs += 1
            begin_last_epoch = m
            epoch_idx += 1
        prev = m

    return obs_epochs, begin_last_epoch



import numpy as np


def X_Z(Z):
    """a function to calculate equation 3 of KS91
       X gives the optical pathlength through the atmosphere at a given
       zenith distance Z (where Z is given in degrees)
    """
    X = np.power(1 - 0.96*np.sin(Z*np.pi/180.)*np.sin(Z*np.pi/180.), -0.5)

    return X


def B_zee(B_zen, kappa, Z):
    """a function to calculate equation 2 of KS91, which gives
       the dark time sky brightness as a function of:

    Parameters:
        ----------

        B_zen: np.float64
            the zenith sky brightness (in nanoLamberts)

        kappa: np.float64
            extinction coefficient (magnitudes per airmass)

        Z: np.float64
            zenith distance of the sight line (in degrees)
    """

    # calculate X_Z via a function call
    X = X_Z(Z)

    # calculate equation 2 of KS91
    B_0 = X*B_zen*np.power(10, -0.4*kappa*(X-1))

    return B_0


def KS91_deltaV(alpha, rho, zee, zee_m, B_zen=79, kappa=0.172):
    """Computer KS91 delta V for a field caused by the moon

    Parameters:
    ----------
    alpha : float, np.float64
        lunar phase angle (degress, 0 is full, 180 is new)
    rho : float, np.float64 (can be array)
        moon/sky seperation (degrees)
    zee : float, np.float64 (can be array)
        zenith distance of the sight line (degrees)
    zee_m : float, np.float64
        zenith distance of the moon (degrees)
    B_zen : float, np.float64
        the zenith sky brightness (in nanoLamberts)
    kappa : float, np.float64
        extinction coefficient (magnitudes per airmass)

    Returns:
    -------
    deltaV : np.float64 (or array)
        change in V mag at targ location due to moon
    """
    # calculate Istar (Eq. 20 in KS91)
    Istar = 10.**(-0.4*(3.84+0.026*np.abs(alpha)+(4e-9)*(alpha**4)))

    # calculate f of rho (eq. 21 in KS91)
    f_rho = (10.**5.36) * (1.06 + (np.cos(rho*(3.1415/180.)))**2.) \
              + np.power(10,  6.15-(rho/40.))

    # calculate B_moon
    B_moon = f_rho * Istar * (10.**(-0.4*kappa*X_Z(zee_m)))\
                   * (1 - np.power(10, -0.4*kappa*X_Z(zee)))

    # calculate deltaV
    deltaV = -2.5 * np.log10((B_moon + B_zee(B_zen, kappa, zee))/B_zee(B_zen, kappa, zee))

    return deltaV



# encoding: utf-8
#
# main.py


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from pytest import mark

import os
import numpy as np
import roboscheduler.cadence as cadence


class TestPacking(object):
    """Tests Packing"""

    def test_set_packing(self):
        clist = cadence.CadenceList()
        clist.reset()
        cfile = os.path.join(os.getenv('ROBOSCHEDULER_DIR'),
                             'data', 'rsc-test-cadence.fits')
        clist.fromfits(cfile)

        packing = cadence.Packing(field_cadence='faint_1x1')
        assert (packing.nepochs == 1)
        for i in np.arange(packing.nepochs, dtype=np.int32):
            assert (packing.epoch_targets[i].max() == -1)
            assert (packing.epoch_targets[i].min() == -1)
            assert (packing.epoch_nexposures[i] == 1)
            assert (packing.epoch_nused[i] == 0)
        for i in np.arange(packing.epoch_nexposures.sum(), dtype=np.int32):
            assert (packing.exposures[i] == -1)

        packing = cadence.Packing(field_cadence='faint_month_4x2')
        assert (packing.nepochs == 4)
        for i in np.arange(packing.nepochs, dtype=np.int32):
            assert (packing.epoch_targets[i].max() == -1)
            assert (packing.epoch_targets[i].min() == -1)
            assert (packing.epoch_nexposures[i] == 2)
            assert (packing.epoch_nused[i] == 0)
        for i in np.arange(packing.epoch_nexposures.sum(), dtype=np.int32):
            assert (packing.exposures[i] == -1)

    def test_check_target(self):
        clist = cadence.CadenceList()
        clist.reset()
        cfile = os.path.join(os.getenv('ROBOSCHEDULER_DIR'),
                             'data', 'rsc-test-cadence.fits')
        clist.fromfits(cfile)

        packing = cadence.Packing(field_cadence='faint_2x4')
        assert (packing.check_target(target_cadence='faint_1x1')['ok']
                is True)
        assert (packing.check_target(target_cadence='faint_month_4x2')['ok']
                is False)
        assert (packing.check_target(target_cadence='faint_2x2')['ok']
                is True)
        assert (packing.check_target(target_cadence='bright_2x2')['ok']
                is True)

        packing = cadence.Packing(field_cadence='bright_month_4x2')
        assert (packing.check_target(target_cadence='bright_1x1')['ok']
                is True)
        assert (packing.check_target(target_cadence='bright_month_4x2')['ok']
                is True)
        assert (packing.check_target(target_cadence='bright_2x2')['ok']
                is True)
        assert (packing.check_target(target_cadence='faint_2x2')['ok']
                is False)
        assert (packing.check_target(target_cadence='bright_2x4')['ok']
                is False)
        assert (packing.check_target(target_cadence='bright_month_2x2')['ok']
                is True)
        assert (packing.check_target(target_cadence='faint_month_2x2')['ok']
                is False)

    # def test_add_target(self):
    #     clist = cadence.CadenceList()
    #     clist.reset()
    #     cfile = os.path.join(os.getenv('ROBOSCHEDULER_DIR'),
    #                          'data', 'rsc-test-cadence.fits')
    #     clist.fromfits(cfile)

    #     packing = cadence.Packing(field_cadence='faint_2x4')
    #     assert(packing.add_target(target_id=1, target_cadence='faint_1x1') is
    #            True)
    #     assert(packing.add_target(target_id=1, target_cadence='faint_2x4') is
    #            False)
    #     assert(packing.epoch_targets[0][0] == 1)
    #     assert(packing.epoch_nused[0] == 1)
    #     assert(packing.epoch_nused.sum() == 1)
    #     assert(packing.exposures[0] == 1)
    #     assert(packing.exposures[1:].max() == -1)

    #     packing = cadence.Packing(field_cadence='faint_2x4')
    #     assert(packing.add_target(target_id=1, target_cadence='faint_2x4') is
    #            True)
    #     assert(packing.add_target(target_id=1, target_cadence='faint_1x4') is
    #            False)
    #     assert(packing.epoch_nused.min() == 4)
    #     assert(packing.epoch_nused.max() == 4)
    #     assert(packing.exposures.min() == 1)
    #     assert(packing.exposures.max() == 1)



# encoding: utf-8
#
# main.py


from __future__ import division
from __future__ import print_function
from __future__ import absolute_import
from __future__ import unicode_literals

from pytest import mark

import os
import roboscheduler.cadence as cadence


class TestCadence(object):
    """Tests CadenceList"""

    def test_cadence_epochs(self):
        clist = cadence.CadenceList()
        clist.reset()
        clist.add_cadence(name='one', nexposures=4,
                          skybrightness=[1., 1., 1., 1.],
                          delta=[0., 0., 1., 0.],
                          delta_min=[0., 0., 0.5, 0.],
                          delta_max=[0., 0., 1.5, 0.],
                          instrument=['boss'] * 4)
        assert clist.cadences['one'].nepochs == 2
        assert clist.cadences['one'].epoch_indx[0] == 0
        assert clist.cadences['one'].epoch_indx[1] == 2

    def test_cadence_single(self):
        clist = cadence.CadenceList()
        clist.reset()
        clist.add_cadence(name='one', nexposures=1,
                          skybrightness=[1.], delta=[-1.], delta_min=[-1.],
                          delta_max=[-1.], instrument=['boss'])
        clist.add_cadence(name='two', nexposures=1,
                          skybrightness=[0.35], delta=[-1.], delta_min=[-1.],
                          delta_max=[-1.], instrument=['boss'])
        assert clist.cadence_consistency('one', 'two', return_solutions=False) == True

    def test_cadence_single_double(self):
        clist = cadence.CadenceList()
        clist.reset()
        clist.add_cadence(name='one', nexposures=1,
                          skybrightness=[1.], delta=[-1.], delta_min=[-1.],
                          delta_max=[-1.], instrument=['boss'])
        clist.add_cadence(name='two', nexposures=2,
                          skybrightness=[0.35, 0.35], delta=[0., 30.],
                          delta_min=[0., 20.],
                          delta_max=[0., 40.],
                          instrument=['boss', 'boss'])
        assert clist.cadence_consistency('one', 'two', return_solutions=False) == True
        assert clist.cadence_consistency('two', 'one', return_solutions=False) == False

    def test_cadence_single_double_skybrightness(self):
        clist = cadence.CadenceList()
        clist.reset()
        clist.add_cadence(name='bright', nexposures=1,
                          skybrightness=[1.], delta=[-1.], delta_min=[-1.],
                          delta_max=[-1.], instrument=['boss'])
        clist.add_cadence(name='dark', nexposures=1,
                          skybrightness=[0.35], delta=[-1.], delta_min=[-1.],
                          delta_max=[-1.], instrument=['boss'])
        clist.add_cadence(name='two', nexposures=2,
                          skybrightness=[1., 1.], delta=[0., 30.],
                          delta_min=[0., 20.],
                          delta_max=[0., 40.],
                          instrument=['boss', 'boss'])
        assert clist.cadence_consistency('bright', 'two',
                                         return_solutions=False) == True
        assert clist.cadence_consistency('dark', 'two',
                                         return_solutions=False) == False

    def test_cadence_double_double(self):
        clist = cadence.CadenceList()
        clist.reset()
        clist.add_cadence(name='one', nexposures=2,
                          skybrightness=[0.35, 0.35], delta=[0., 30.],
                          delta_min=[0., 19.],
                          delta_max=[0., 41.],
                          instrument=['boss', 'boss'])
        clist.add_cadence(name='two', nexposures=2,
                          skybrightness=[0.35, 0.35], delta=[0., 30.],
                          delta_min=[0., 20.],
                          delta_max=[0., 40.],
                          instrument=['boss', 'boss'])
        assert clist.cadence_consistency('one', 'two',
                                         return_solutions=False) is True
        assert clist.cadence_consistency('two', 'one',
                                         return_solutions=False) is False

    def test_cadence_file(self):
        clist = cadence.CadenceList()
        clist.reset()
        cfile = os.path.join(os.getenv('ROBOSCHEDULER_DIR'),
                             'data', 'rsc-test-cadence.fits')
        clist.fromfits(cfile)

        assert clist.ncadences == 12

        assert (clist.cadence_consistency('bright_1x1', 'faint_1x1',
                                          return_solutions=False) is True)

        assert (clist.cadence_consistency('faint_1x1', 'bright_1x1',
                                          return_solutions=False) is False)

        assert (clist.cadence_consistency('faint_1x1', 'faint_1x4',
                                          return_solutions=False) is True)

        assert (clist.cadence_consistency('faint_1x4', 'faint_1x1',
                                          return_solutions=False) is False)

        assert (clist.cadence_consistency('faint_1x4', 'faint_1x4',
                                          return_solutions=False) is True)

        assert (clist.cadence_consistency('faint_1x4', 'faint_1x4',
                                          return_solutions=False) is True)

        assert (clist.cadence_consistency('faint_2x2', 'faint_month_2x2',
                                          return_solutions=False) is True)

        assert (clist.cadence_consistency('faint_2x2', 'faint_month_4x2',
                                          return_solutions=False) is True)

        assert (clist.cadence_consistency('faint_month_2x2', 'faint_month_4x2',
                                          return_solutions=False) is True)

        assert (clist.cadence_consistency('faint_month_2x2', 'faint_month_4x2',
                                          return_solutions=False) is True)

        assert (clist.cadence_consistency('faint_1x4', 'faint_month_4x2',
                                          return_solutions=False) is False)



from __future__ import print_function, division
import numpy as np
from PyAstronomy.pyasl import moonpos
from roboscheduler.sunpos2 import sunpos2


def moonphase2(jd):
  """
    Computes the illuminated fraction of the Moon at given Julian date(s).
    
    Parameters
    ----------
    jd : float or array
         The Julian date.
         
    Returns
    -------
    Fraction : float or array
        The illuminated fraction [0 - 1] of the Moon.
        Has the same size as `jd`.

    Notes
    -----
    
    .. note:: This function was ported from the IDL Astronomy User's Library.

    :IDL - Documentation:  
    
    NAME:
          MPHASE
    PURPOSE:
          Return the illuminated fraction of the Moon at given Julian date(s) 
    
    CALLING SEQUENCE:
          MPHASE, jd, k
    INPUT:
          JD - Julian date, scalar or vector, double precision recommended
    OUTPUT:
          k - illuminated fraction of Moon's disk (0.0 < k < 1.0), same number
              of elements as jd.   k = 0 indicates a new moon, while k = 1 for
              a full moon.
    EXAMPLE:
          Plot the illuminated fraction of the moon for every day in July 
          1996 at 0 TD (~Greenwich noon).
    
          IDL> jdcnv, 1996, 7, 1, 0, jd         ;Get Julian date of July 1
          IDL> mphase, jd+dindgen(31), k        ;Moon phase for all 31 days
          IDL> plot, indgen(31),k               ;Plot phase vs. July day number
  """

  jd = np.array(jd, ndmin=1)
  
  # Earth-Sun distance (1 AU)
  edist = 1.49598e8         
  
  mpos = moonpos(jd)
  ram = mpos[0]*np.pi/180.
  decm = mpos[1]*np.pi/180.
  dism = mpos[2]
  
  spos = sunpos2(jd)
  ras = spos[1]*np.pi/180.
  decs = spos[2]*np.pi/180.

  phi = np.arccos( np.sin(decs)*np.sin(decm) + np.cos(decs)*np.cos(decm)*np.cos(ras-ram) )
  inc = np.arctan2( edist * np.sin(phi), dism - edist*np.cos(phi) )
  k = (1 + np.cos(inc))/2.
  
  return np.ravel(k)



"""
 $Id:$
 $Rev::                                  $:  # Revision of last commit.
 $LastChangedBy::                        $:  # Author of last commit.
 $LastChangedDate::                      $:  # Date of last commit.


 Taken from Erin Sheldon's esutil package at:
 http://esutil.googlecode.com/svn/trunk/esutil/coords.py
 Felipe Menanteau, NCSA, March 2014.

    NAME
        coords
    PURPOSE
        A set of astronomical utilities for dealing with coordinates and
        coordinate transformations.

    COORDINATE TRANSFORMATIONS
        euler:  
            A generic routine for transforming between Galactic, Celestial,
            and ecliptic coords.  The following wrapper routines are also
            supplied for convenience:

        l,b = eq2gal(ra, dec, b1950=False, dtype='f8')
            Convert equatorial to glactic coordinates.

        # The following use the same interface:
        gal2eq
            Convert galactic to equatorial coordinates.
        eq2ec
            Convert equatorial to ecliptic coordinates.
        ec2eq
            Convert ecliptic to equatorial coordinates.
        ec2gal
            Convert ecliptic to galactic coordinates.
        gal2ec
            Convert galactic to ecliptic coordinates.

        # These SDSS specific functions do not use euler
        eq2sdss
            Convert between equatorial and corrected SDSS survey coords.
        sdss2eq
            Convert between corrected SDSS survey and equatorial coords.

        eq2xyz: Convert equatorial to x,y,z on the sphere according to
            the following transform:
                    x = sin(pi/2-dec)*cos(ra)
                    y = sin(pi/2-dec)*sin(ra)
                    z = cos(pi/2-dec)

        xyz2eq: 
            inverse of eq2xyz

        sphdist:
            Calculate the arc length between two sets of points on the sphere.
            Currently only takes ra,dec.

        shiftlon:
            shift the input longitude.  By default wrap the coordinate to
            -180,180.  If a shift is entered, return the new value
            lon-shift such that the range is still [0,360)
            
        shiftra:
            shift right ascension.  This just calls shiftlon

        radec2aitoff:
            Convert ra,dec to aitoff coordinates.

        dec_parse(decstring)
            parse a colon separated string representing declination ito
            degrees.
        ra_parse(decstring)
            parse a colon separated string representing right ascension ito
            degrees.

        randsphere(numrand, system='eq', ra_range=[0,360], dec_range=[-90,90]):
            Generate random points on the sphere.  By default ra,dec are
            returned.  If system='xyz' then x,y,z are returned.

        randcap(nrand,ra,dec,rad,get_radius=False):
            Create random points in a cap, or disc, centered at the
            input ra,dec location and with radius rad.

        rect_area(lon_min, lon_max, lat_min, lat_max)
            Calculate the area of a rectangle on the sphere.


"""
license="""
  Copyright (C) 2009  Erin Sheldon

    This program is free software; you can redistribute it and/or modify it
    under the terms of version 2 of the GNU General Public License as
    published by the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

"""

try:
    import numpy
    from numpy import where, sin, cos, arccos, arcsin, arctan2, sqrt, rad2deg, deg2rad
    have_numpy=True
except:
    have_numpy=False

import math
PI=math.pi
HALFPI = PI/2.0
D2R = PI/180.0
R2D = 1.0/D2R

_sdsspar = {}
_sdsspar['center_ra'] = 185.0
_sdsspar['center_dec'] = 32.5
_sdsspar['node'] = (_sdsspar['center_ra'] - 90.0)*D2R
_sdsspar['etapole'] = _sdsspar['center_dec']*D2R
_sdsspar['etaoffset'] = 91.25

_sdsspar['doc'] = """
    A set of transformation functions for use with SDSS coordinate systems.

    eq2sdss(): Convert between equatorial and corrected SDSS survey coords.
    sdss2eq(): Convert between corrected SDSS survey and equatorial coords.

    Don't use these unless you have to, as these are the old coordinates
        eq2survey(): Convert between equatorial and uncorrected survey coords.
        survey2eq(): Convert between uncorrected survey and equatorial coords.

    Adapted from astrotools
        Erin Sheldon, NYU, 2006-03-11
    Force data type and allow selection of dtype through keyword.
        Erin Sheldon, NYU, 2007-05-23

"""


def euler(ai_in, bi_in, select, b1950=False, dtype='f8'):
    """
    NAME:
        euler
    PURPOSE:
        Transform between Galactic, celestial, and ecliptic coordinates.

    CALLING SEQUENCE:
        long_out, lat_out = 
            euler(long_in, lat_in, type, b1950=False, dtype='f8')

    INPUTS:
       long_in - Input Longitude in DEGREES, scalar or vector.  
       lat_in  - Input Latitude in DEGREES
       select  - Integer (1-6) specifying type of coordinate transformation.  

      select   From          To        |   select      From            To
       1     RA-Dec (2000)  Galactic   |     4       Ecliptic      RA-Dec    
       2     Galactic       RA-DEC     |     5       Ecliptic      Galactic  
       3     RA-Dec         Ecliptic   |     6       Galactic      Ecliptic  

      Celestial coordinates (RA, Dec) should be given in equinox J2000 
      unless the b1950=True keyword is set.

    OUTPUTS:
       long_out - Output Longitude in DEGREES
       lat_out  - Output Latitude in DEGREES

    INPUT KEYWORD:
       b1950 - If this keyword is true then input and output 
             celestial and ecliptic coordinates should be given in equinox 
             B1950.
    REVISION HISTORY:
       Written W. Landsman,  February 1987
       Adapted from Fortran by Daryl Yentis NRL
       Converted to IDL V5.0   W. Landsman   September 1997
       Made J2000 the default, added /FK4 keyword  W. Landsman December 1998
       Add option to specify SELECT as a keyword W. Landsman March 2003

       Converted from IDL to numerical Python: Erin Sheldon, NYU, 2008-07-02

    """

    # Make a copy as an array. ndmin=1 to avoid messed up scalar arrays
    ai = numpy.array(ai_in, ndmin=1, copy=True, dtype=dtype)
    bi = numpy.array(bi_in, ndmin=1, copy=True, dtype=dtype)

    twopi   =   2.0*PI
    fourpi  =   4.0*PI

    #   J2000 coordinate conversions are based on the following constants
    #   (see the Hipparcos explanatory supplement).
    #  eps = 23.4392911111d           Obliquity of the ecliptic
    #  alphaG = 192.85948d            Right Ascension of Galactic North Pole
    #  deltaG = 27.12825d             Declination of Galactic North Pole
    #  lomega = 32.93192d             Galactic longitude of celestial equator  
    #  alphaE = 180.02322d            Ecliptic longitude of Galactic North Pole
    #  deltaE = 29.811438523d         Ecliptic latitude of Galactic North Pole
    #  Eomega  = 6.3839743d           Galactic longitude of ecliptic equator              
    # Parameters for all the different conversions
    if b1950:

        equinox = '(B1950)' 
        psi    = numpy.array([ 0.57595865315, 4.9261918136,  
                              0.00000000000, 0.0000000000,  
                              0.11129056012, 4.7005372834], dtype=dtype)
        stheta = numpy.array([ 0.88781538514,-0.88781538514, 
                              0.39788119938,-0.39788119938, 
                              0.86766174755,-0.86766174755], dtype=dtype)
        ctheta = numpy.array([ 0.46019978478, 0.46019978478, 
                              0.91743694670, 0.91743694670, 
                              0.49715499774, 0.49715499774], dtype=dtype) 
        phi  =   numpy.array([ 4.9261918136,  0.57595865315, 
                              0.0000000000, 0.00000000000, 
                              4.7005372834, 0.11129056012], dtype=dtype)

    else:

        equinox = '(J2000)'

        psi    = numpy.array([ 0.57477043300, 4.9368292465,  
                              0.00000000000, 0.0000000000,    
                              0.11142137093, 4.71279419371], dtype=dtype)
        stheta = numpy.array([ 0.88998808748,-0.88998808748, 
                              0.39777715593,-0.39777715593, 
                              0.86766622025,-0.86766622025], dtype=dtype)
        ctheta = numpy.array([ 0.45598377618, 0.45598377618, 
                              0.91748206207, 0.91748206207, 
                              0.49714719172, 0.49714719172], dtype=dtype)
        phi    = numpy.array([ 4.9368292465,  0.57477043300, 
                              0.0000000000, 0.00000000000, 
                              4.71279419371, 0.11142137093], dtype=dtype)

    # zero offset
    i  = select - 1 
    a  = ai*D2R - phi[i]
    
    b = bi*D2R
    sb = sin(b)
    cb = cos(b)
    cbsa = cb * sin(a)
    b  = -stheta[i] * cbsa + ctheta[i] * sb
    w, = numpy.where(b > 1.0)
    if w.size > 0:
        b[w] = 1.0
    bo    = arcsin(b)*R2D

    a =  arctan2( ctheta[i] * cbsa + stheta[i] * sb, cb * cos(a) )

    ao = ( (a+psi[i]+fourpi) % twopi) * R2D

    return ao, bo



#
# Some clearer shortcut functions which call Euler
#
def eq2gal(ra, dec, b1950=False, dtype='f8'):
    """
    NAME
        eq2gal
    PURPOSE
        Convert from equatorial to galactic coordinates in units of degrees.
    CALLING SEQUENCE
        l,b = eq2gal(ra, dec, b1950=False, dtype='f8')
    INPUTS
        ra, dec: Equatorial coordinates.  May be Numpy arrays, sequences, or
            scalars as long as they are all the same length.  They must be
            convertible to a Numpy array with the specified datatype.
    KEYWORDS
        b1950:  If True, use b1950 coordiates.  By default j2000 are used.
        dtype:  The datatype of the output arrays.  Default is f8
    OUTPUTS
        l, b:  Galactic longitude and latitude.  The returned value is always
            a Numpy array with the specified dtype
    REVISION HISTORY
        Created Erin Sheldon, NYU, 2008-07-02
    """
    return euler(ra, dec, 1, b1950=b1950, dtype=dtype)


def gal2eq(l, b, b1950=False, dtype='f8'):
    """
    NAME
        gal2eq
    PURPOSE
        Convert from galactice to equatorial coordinates in units of degrees.
    CALLING SEQUENCE
        ra,dec = gal2eq(l, b, b1950=False, dtype='f8')
    INPUTS
        l, b: Galactic coordinates.  May be Numpy arrays, sequences, or
            scalars as long as they are all the same length.  They must be
            convertible to a Numpy array with the specified datatype.
    KEYWORDS
        b1950:  If True, use b1950 coordiates.  By default j2000 are used.
        dtype:  The datatype of the output arrays.  Default is f8
    OUTPUTS
        ra, dec:  Equatorial longitude and latitude.  The returned value is 
            always a Numpy array with the specified dtype
    REVISION HISTORY
        Created Erin Sheldon, NYU, 2008-07-02
    """

    return euler(l, b, 2, b1950=b1950, dtype=dtype)

def eq2ec(ra, dec, b1950=False, dtype='f8'):
    """
    NAME
        eq2ec
    PURPOSE
        Convert from equatorial to ecliptic coordinates in units of degrees.
    CALLING SEQUENCE
        lam,beta = eq2ec(ra, dec, b1950=False, dtype='f8')
    INPUTS
        ra, dec: Equatorial coordinates.  May be Numpy arrays, sequences, or
            scalars as long as they are all the same length.  They must be
            convertible to a Numpy array with the specified datatype.
    KEYWORDS
        b1950:  If True, use b1950 coordiates.  By default j2000 are used.
        dtype:  The datatype of the output arrays.  Default is f8
    OUTPUTS
        lam, beta:  Ecliptic longitude and latitude.  The returned value is 
            always a Numpy array with the specified dtype
    REVISION HISTORY
        Created Erin Sheldon, NYU, 2008-07-02
    """

    return euler(ra, dec, 3, b1950=b1950, dtype=dtype)

def ec2eq(lam, beta, b1950=False, dtype='f8'):
    """
    NAME
        ec2eq
    PURPOSE
        Convert from ecliptic to equatorial coordinates in units of degrees.
    CALLING SEQUENCE
        ra,dec = eq2gal(lam, beta, b1950=False, dtype='f8')
    INPUTS
        lam,beta: Ecliptic coordinates.  May be Numpy arrays, sequences, or
            scalars as long as they are all the same length.  They must be
            convertible to a Numpy array with the specified datatype.
    KEYWORDS
        b1950:  If True, use b1950 coordiates.  By default j2000 are used.
        dtype:  The datatype of the output arrays.  Default is f8
    OUTPUTS
        ra,dec:  Equatorial longitude and latitude.  The returned value is 
            always a Numpy array with the specified dtype
    REVISION HISTORY
        Created Erin Sheldon, NYU, 2008-07-02
    """

    return euler(lam, beta, 4, b1950=b1950, dtype=dtype)

def ec2gal(lam, beta, b1950=False, dtype='f8'):
    """
    NAME
        ec2gal
    PURPOSE
        Convert from ecliptic to galactic coordinates in units of degrees.
    CALLING SEQUENCE
        l,b = eq2gal(lam, beta, b1950=False, dtype='f8')
    INPUTS
        lam, beta: Ecliptic coordinates.  May be Numpy arrays, sequences, or
            scalars as long as they are all the same length.  They must be
            convertible to a Numpy array with the specified datatype.
    KEYWORDS
        b1950:  If True, use b1950 coordiates.  By default j2000 are used.
        dtype:  The datatype of the output arrays.  Default is f8
    OUTPUTS
        l, b:  Galactic longitude and latitude.  The returned value is always
            a Numpy array with the specified dtype
    REVISION HISTORY
        Created Erin Sheldon, NYU, 2008-07-02
    """

    return euler(lam, beta, 5, b1950=b1950, dtype=dtype)

def gal2ec(l, b, b1950=False, dtype='f8'):
    """
    NAME
        gal2ec
    PURPOSE
        Convert from Galactic to Ecliptic coordinates in units of degrees.
    CALLING SEQUENCE
        lam,beta = eq2gal(l, b, b1950=False, dtype='f8')
    INPUTS
        l, b: Galactic coordinates.  May be Numpy arrays, sequences, or
            scalars as long as they are all the same length.  They must be
            convertible to a Numpy array with the specified datatype.
    KEYWORDS
        b1950:  If True, use b1950 coordiates.  By default j2000 are used.
        dtype:  The datatype of the output arrays.  Default is f8
    OUTPUTS
        lam,beta:  Ecliptic longitude and latitude.  The returned value is 
            always a Numpy array with the specified dtype
    REVISION HISTORY
        Created Erin Sheldon, NYU, 2008-07-02
    """

    return euler(l, b, 6, b1950=b1950, dtype=dtype)


def _thetaphi2xyz(theta, phi):
    """
    theta and phi in radians relative to the SDSS node at ra=95 degrees
    """
    x = cos(theta)*cos(phi)
    y = sin(theta)*cos(phi)
    z = sin(phi)

    return x,y,z

def _xyz2thetaphi(x,y,z):
    """
    returns theta, phi in radians relative to the SDSS node at ra=95 degrees
    """
    phi = arcsin(z)
    theta = arctan2(y,x)

    return theta, phi

def eq2xyz(ra, dec, dtype='f8', units='deg'):
    """
    Convert equatorial coordinates RA and DEC to x,y,z on the unit sphere

    parameters
    ----------
    ra: scalar or array
        Right ascension. Can be an array
    dec: scalar or array
        Declination. Can be an array
    units: string, optional
        'deg' if the input is degrees, 'rad' if input
        is in radians.  Default is degrees.

    Notes:
        This follows the same convention as the STOMP package.
    """

    theta = numpy.array(ra, ndmin=1, copy=True, dtype=dtype)
    phi = numpy.array(dec, ndmin=1, copy=True, dtype=dtype)

    # in place is more efficient
    if units == 'deg':
        numpy.deg2rad(theta,theta)
        numpy.deg2rad(phi,phi)

    theta -= _sdsspar['node']

    return _thetaphi2xyz(theta, phi)

def xyz2eq(xin,yin,zin, units='deg'):
    """
    Convert x,y,z on the unit sphere to RA DEC.

    parameters
    ----------
    x,y,z: 
        scalars or arrays as given by eq2xyz
    units: string, optional
        'deg' if the output is to be degrees, 'rad' if it is to be radians.
        Default is degrees.

    Notes:
        This follows the same convention as the STOMP package.
    """

    x = numpy.array(xin, ndmin=1, copy=False)
    y = numpy.array(yin, ndmin=1, copy=False)
    z = numpy.array(zin, ndmin=1, copy=False)

    theta,phi = _xyz2thetaphi(x,y,z)
    theta += _sdsspar['node']

    if units == 'deg':
        numpy.rad2deg(theta,theta)
        numpy.rad2deg(phi,phi)

    atbound(theta, 0.0, 360.0)

    # theta->ra, phi->dec
    return theta,phi

   

def sphdist(ra1, dec1, ra2, dec2, units=['deg','deg']):
    """
    Get the arc length between two points on the unit sphere

    parameters
    ----------
    ra1,dec1,ra2,dec2: scalar or array
        Coordinates of two points or sets of points. 
        Must be the same length.
    units: sequence
        A sequence containing the units of the input and output.  Default
        ['deg',deg'], which means inputs and outputs are in degrees.  Units
        can be 'deg' or 'rad'
    """
    
    units_in,units_out = units

    # note x,y,z from eq2xyz always returns 8-byte float
    x1,y1,z1 = eq2xyz(ra1, dec1, units=units_in)
    x2,y2,z2 = eq2xyz(ra2, dec2, units=units_in)

    costheta = x1*x2 + y1*y2 + z1*z2
    costheta.clip(-1.0,1.0,out=costheta)

    theta = arccos(costheta)

    if units_out == 'deg':
        numpy.rad2deg(theta,theta)
    return theta


def gcirc(ra1deg,dec1deg,ra2deg,dec2deg,getangle=False):
    """
    This is currently very inflexible: degrees in, radians out
    """
    ra1  = numpy.array(ra1deg, dtype='f8',ndmin=1)
    dec1 = numpy.array(dec1deg,dtype='f8',ndmin=1)
    ra2  = numpy.array(ra2deg, dtype='f8',ndmin=1)
    dec2 = numpy.array(dec2deg,dtype='f8',ndmin=1)

    deg2rad(ra1,ra1)
    deg2rad(dec1,dec1)
    deg2rad(ra2,ra2)
    deg2rad(dec2,dec2)

    sindec1 = sin(dec1)
    cosdec1 = cos(dec1)

    sindec2 = sin(dec2)
    cosdec2 = cos(dec2)

    radiff = (ra2-ra1)
    cosradiff = cos(radiff)
    cosdis = sindec1*sindec2 + cosdec1*cosdec2*cosradiff

    cosdis.clip(-1.0,1.0,out=cosdis)
    dis = arccos(cosdis)

    if getangle:
        theta = arctan2( sin(radiff), 
                         (sindec1*cosradiff - cosdec1*sindec2/cosdec2) ) - HALFPI
        return dis,theta
    else:
        return dis

# utility functions
def atbound(longitude, minval, maxval):
    w, = numpy.where(longitude < minval)
    while w.size > 0:
        longitude[w] += 360.0
        w, = numpy.where( longitude < minval )

    w, = numpy.where(longitude > maxval)
    while w.size > 0:
        longitude[w] -= 360.0
        w, = numpy.where( longitude > maxval )

    return

def atbound2(theta, phi):

    atbound(theta, -180.0, 180.0)

    w, = numpy.where( numpy.abs(theta) > 90.0 )
    if w.size > 0:
        theta[w] = 180.0 - theta[w]
        phi[w] += 180.0

    atbound(theta, -180.0, 180.0)
    atbound(phi, 0.0, 360.0)

    w, = numpy.where( numpy.abs(theta) == 90.0 )
    if w.size > 0:
        phi[w] = 0.0
           
#
# SDSS specific conversions
#


def eq2sdss(ra_in, dec_in, dtype='f8'):
    """
    NAME:
      eq2sdss
    PURPOSE:
       Convert from ra, dec to the corrected clambda, ceta 
       SDSS survey coordinate system.  It is corrected so that the
       longitude eta ranges from [-180.0, 180.0] and the latitude
       lambda ranges from [-90.0,90.0].  The standard lambda/eta 
       both range from [-180.0,180.0] which doesn't make sense.
       NOTE: lambda is often referred to as longitude but this
       is incorrect since it has poles at [-90,90]

    CALLING SEQUENCE:
      from esutil import coords
      (clambda, ceta) = coords.eq2sdss(ra, dec, dtype='f8')

    INPUTS: 
      ra: Equatorial latitude in degrees. 
      dec: Equatorial longitude in degrees. 
    OPTIONAL INPUTS:
        dtype: The data type of output.  Default is 'f8'. See 
        numpy.typeDict for a list of possible types.
        dtype: The data type of output.  Default is 'f8'.

    OUTPUTS: 
      clambda: Corrected Survey longitude (actually lattitude) in degrees
      ceta: Corrected Survey latitude (actually logitude) in degrees
      
    REVISION HISTORY:
      Written: 11-March-2006  Converted from IDL program.
    """

    # Make a copy as an array. ndmin=1 to avoid messed up scalar arrays
    ra = numpy.array(ra_in, ndmin=1, copy=True, dtype=dtype)
    dec = numpy.array(dec_in, ndmin=1, copy=True, dtype=dtype)

    if (ra.size != dec.size):
        raise ValueError("RA, DEC must be same size")

    # range checking
    if (ra.min() < 0.0) | (ra.max() > 360.0):
        raise ValueError('RA must we within [0,360]')
    if (dec.min() < -90.0) | (dec.max() > 90.0):
        raise ValueError('DEC must we within [-90,90]')

    ra *= D2R
    dec *= D2R
    ra -= _sdsspar['node']

    # generate x,y,z on unit sphere, clearing memory as we go
    cdec = cos(dec)
    
    x = cos(ra)*cdec
    y = sin(ra)*cdec

    ra = 0; cdec = 0 # mem 
    
    z = numpy.sin(dec)
    
    dec = 0 # mem

    # generate clambda, ceta
    # do things in place to save memory

    # clambda = -arcsin( x ) (not a copy clambda=x)
    arcsin(x, x); clambda=x
    clambda *= -1

    
    arctan2( z, y, z); ceta = z
    ceta -= _sdsspar['etapole']

    clambda *= R2D
    ceta *= R2D

    atbound(ceta, -180.0, 180.0)
    
    return (clambda, ceta)

def sdss2eq(clambda_in, ceta_in, dtype='f8'):
    """
    NAME:
      sdss2eq
    PURPOSE:
       Convert corrected clambda, ceta SDSS survey coordinate system t
       equatorial coords.  

    CALLING SEQUENCE:
      from esutil import coords
      (ra, dec) = coords.sdss2eq(clambda, ceta, dtype='f8')

    INPUTS: 
      clambda: Corrected Survey longitude (actually lattitude) in degrees
      ceta: Corrected Survey latitude (actually logitude) in degrees
    OPTIONAL INPUTS:
        dtype: The data type of output.  Default is 'f8'. See 
        numpy.typeDict for a list of possible types.

    OUTPUTS: 
      ra: Equatorial latitude in degrees. 
      dec: Equatorial longitude in degrees. 
      
    REVISION HISTORY:
      Written: 11-March-2006  Converted from IDL program.
    """
    
    # Make a copy as an array. ndmin=1 to avoid messed up scalar arrays
    clambda = numpy.array(clambda_in, ndmin=1, copy=True, dtype=dtype)
    ceta = numpy.array(ceta_in, ndmin=1, copy=True, dtype=dtype)

    # range checking
    if (clambda.min() < -90.0) | (clambda.max() > 90.0):
        raise ValueError('CLAMBDA must we within [-90,90]')
    if (ceta.min() < -180.0) | (ceta.max() > 180.0):
        raise ValueError('CETA must we within [-180,180]')

    clambda *= D2R
    ceta    *= D2R

    x = -sin(clambda)
    y = cos(ceta + _sdsspar['etapole'])*cos(clambda)
    z = sin(ceta + _sdsspar['etapole'])*cos(clambda)
    
    ra = arctan2( y, x ) + _sdsspar['node']
    dec = arcsin(z)

    ra  *= R2D
    dec *= R2D
    atbound2(dec, ra)

    return (ra,dec)


def _eq2survey(ra_in, dec_in, dtype='f8'):
    """
    NAME:
      _eq2survey
    PURPOSE:
       Convert from ra, dec to the lambda, eta 
       SDSS survey coordinate system.  Note this coordinate system is
       not well defined.  Recommend you use csurvey coords.

    CALLING SEQUENCE:
      from esutil import coords
      (lambda, eta) = coords._eq2survey(ra, dec, dtype='f8')

    INPUTS: 
      ra: Equatorial latitude in degrees. 
      dec: Equatorial longitude in degrees. 
    OPTIONAL INPUTS:
        dtype: The data type of output.  Default is 'f8'. See 
        numpy.typeDict for a list of possible types.

    OUTPUTS: 
      lambda: SDSS Survey longitude (actually lattitude) in degrees
      eta: SDSS Survey latitude (actually logitude) in degrees
      
    REVISION HISTORY:
      Written: 11-March-2006  Converted from IDL program.
    """

    # Make a copy as an array. ndmin=1 to avoid messed up scalar arrays
    ra = numpy.array(ra_in, ndmin=1, copy=True, dtype=dtype)
    dec = numpy.array(dec_in, ndmin=1, copy=True, dtype=dtype)

    if (ra.size != dec.size):
        raise ValueError("RA, DEC must be same size")

    # range checking
    if (ra.min() < 0.0) | (ra.max() > 360.0):
        raise ValueError('RA must we within [0,360]')
    if (dec.min() < -90.0) | (dec.max() > 90.0):
        raise ValueError('DEC must we within [-90,90]')



    ra *= D2R
    dec *= D2R
    ra -= _sdsspar['node']

    # generate x,y,z on unit sphere, clearing memory as we go
    cdec = cos(dec)
    
    x = cos(ra)*cdec
    y = sin(ra)*cdec

    ra = 0; cdec = 0 # mem 
    
    z = sin(dec)
    
    dec = 0 # mem

    # generate lam, eta
    # do things in place to save memory

    # lam = -arcsin( x ) (not a copy lam=x)
    arcsin(x, x); lam=x
    lam *= -1

    
    arctan2( z, y, z); eta = z
    eta -= _sdsspar['etapole']

    lam *= R2D
    eta *= R2D

    atbound2(lam, eta)
    atbound(eta, -180.0, 180.0)

    w, = numpy.where( eta > (90.0 - _sdsspar['center_dec']) )
    if w.size > 0:
        eta[w] -= 180.0
        lam[w]  = 180.0 - lam[w]

    atbound(lam, -180.0, 180.0)

    return (lam, eta)


def _survey2eq(ra, dec, dtype='f8'):
    """
    NAME:
      _survey2eq
    PURPOSE:
       Convert clambda, ceta SDSS survey coordinate system to
       equatorial coords.  

    CALLING SEQUENCE:
      from esutil import coords
      (ra, dec) = coords._survey2eq(lam, eta, dtype='f8')

    INPUTS: 
      lambda: Survey longitude (actually lattitude) in degrees
      eta:    Survey latitude (actually logitude) in degrees
    OPTIONAL INPUTS:
        dtype: The data type of output.  Default is 'f8'. See 
        numpy.typeDict for a list of possible types.
      
    OUTPUTS: 
      ra: Equatorial latitude in degrees. 
      dec: Equatorial longitude in degrees. 
      
    REVISION HISTORY:
      Written: 11-March-2006  Converted from IDL program.
    """

    return csurvey2eq(ra,dec, dtype=dtype)


def dec_parse(decstring):
    """
    dec = dec_parse(decstring)

    parse a colon separated string representing declination ito
    degrees.
    """
    dec = 0.0

    ds = decstring.split(':')
    lds = len(ds)
    if lds >= 1:
        deg = float(ds[0])
        dec += deg
    if lds >= 2:
        minutes = float(ds[1])
        dec += minutes/60.0
    if lds >= 3:
        sec = float(ds[2])
        dec += sec/3600.0
    return dec

def ra_parse(rastring, hours=True):
    """
    ra = ra_parse(decstring)

    parse a colon separated string representing right ascension ito
    degrees.
    """
    ra = 0.0

    rs = rastring.split(':')
    lrs = len(rs)
    if lrs >= 1:
        deg = float(rs[0])
        ra += deg
        if hours:
            ra *= 15
    if lrs >= 2:
        minutes = float(rs[1])
        ra += minutes/60.0
    if lrs >= 3:
        sec = float(rs[2])
        ra += sec/3600.0
    return ra


def fitsheader2dict(hdr, ext=0):
    """
    Convert a fits header object into a dict.  A dict provides more expected
    interface to the data but cannot be written back to a fits file without
    transformation.
    """

    hdict={}
    for key in hdr:
        hdict[key.lower()] = hdr[key]

    return hdict



def shiftlon(lon_input, shift=None, wrap=True):
    """
    Name:
        shiftlon
    Calling Sequence:
        newlon = shiftlon(longitude, wrap=True, shift=0.0)

    Purpose:

        Shift the value of a longitude.  By default, the value is "wrapped" to
        be [-180,180] instead of [0,360]

        If the shift keyword is sent, then the longitude is simply shifted by
        the input value and then constrained to be again on the [0,360) range.
    
    Input:
        A longitude or array of longitudes on the range [0,360)

    Keywords:
        shift: 
            If shift is sent, then lon-shift is returned, constrained to still
            be on [0,360).
        
        wrap: 
            If shift is not sent, and wrap is True, wrap the range to
            [-180,180]

    """
    lon = numpy.array(lon_input, ndmin=1, copy=True, dtype='f8')

    if shift is not None:
        negshift = False
        if shift < 0:
            negshift=True

        abs_shift = abs(shift)

        # make sure in range [0,360)
        abs_shift = abs_shift % 360.0

        if negshift:
            lon += abs_shift

            w, = numpy.where(lon > 360.0)
            if w.size > 0:
                lon[w] -= 360.0
        else:
            lon -= abs_shift

            w, = numpy.where(lon < 0.0)
            if w.size > 0:
                lon[w] += 360.0

    elif wrap:
        w,=where(lon > 180)
        if w.size > 0:
            lon[w] -= 360

    return lon



def shiftra(ra, shift=None, wrap=True):
    """
    Name:
        shiftra
    Calling Sequence:
        newra = shiftra(ra, wrap=True, shift=0.0)

    Purpose:

        Shift the value of a longitude RA.  By default, the value is "wrapped"
        to be [-180,180] instead of [0,360]

        If the shift keyword is sent, then the longitude is simply shifted by
        the input value and then constrained to be again on the [0,360) range.
    
    Input:
        ra or any other longitude on the range [0,360)

    Keywords:
        shift: 

            If shift is sent, then ra-shift is returned, constrained to still
            be on [0,360).
        
        wrap: 
            If shift is not sent, and wrap is True, wrap the range to
            [-180,180]

    """
    return shiftlon(ra, shift=shift, wrap=wrap)



def radec2aitoff(ra, dec):
    """
    Take the ra/dec into aitoff coords
    """

    r2 = numpy.sqrt(2.0)
    f = 2.*r2/PI


    sra = shiftra(ra)

    alpha2 = sra/2.*D2R
    delta = dec*D2R

    cdec = cos(delta)
    
    denom = sqrt(1.0 + cdec*cos(alpha2))

    x = cdec*sin(alpha2)*2.*r2/denom
    y = sin(delta)*r2/denom
    x = x*R2D/f
    y = y*R2D/f

    crap="""
    sa = l
    if N_elements(sa) eq 1 then sa = fltarr(1) + sa
    x180 = where (sa gt 180.0)
    if x180[0] ne -1 then sa[x180]  = sa[x180] - 360.
    alpha2 = sa/(2*!RADEG)
    delta = b/!RADEG   
    r2 = sqrt(2.)    
    f = 2*r2/!PI     
    cdec = cos(delta)    
    denom =sqrt(1. + cdec*cos(alpha2))
    x = cdec*sin(alpha2)*2.*r2/denom
    y = sin(delta)*r2/denom
    x = x*!radeg/f
    y = y*!radeg/f
    """

    return x,y

def _check_range(rng, allowed):
    if rng is None:
        rng = allowed
    else:
        if not hasattr(rng,'__len__'):
            raise ValueError("range object does not have len() method")

        if rng[0] < allowed[0] or rng[1] > allowed[1]:
            raise ValueError("lon_range should be within [%s,%s]" % allowed)
    return rng

def randsphere(num, ra_range=None, dec_range=None, system='eq'):
    """
    Generate random points on the sphere

    You can limit the range in ra and dec.  To generate on a spherical cap, see
    randcap() 

    parameters
    ----------
    num: integer 
        The number of randoms to generate
    ra_range: list, optional
        Should be within range [0,360].  Default [0,360]
    dec_range: list, optional
        Should be within range [-90,90].  Default [-90,90]
    system: string
        Default is 'eq' for the ra-dec system.  Can also be 'xyz'.

    output
    ------
        for system == 'eq' the return is a tuple
            ra,dec = randsphere(...)
        for system == 'xyz' the return is a tuple
            x,y,z = randsphere(...)

    examples
    --------
        ra,dec = randsphere(2000, ra_range=[10,35], dec_range=[-25,15])
        x,y,z = randsphere(2000, system='xyz')

    """

    ra_range = _check_range(ra_range, [0.0,360.0])
    dec_range = _check_range(dec_range, [-90.0,90.0])

    ra = numpy.random.random(num)
    ra *= (ra_range[1]-ra_range[0])
    if ra_range[0] > 0:
        ra += ra_range[0]

    # number [-1,1)
    cosdec_min = cos(deg2rad(90.0+dec_range[0]))
    cosdec_max = cos(deg2rad(90.0+dec_range[1]))
    v = numpy.random.random(num)
    v *= (cosdec_max-cosdec_min)
    v += cosdec_min

    numpy.clip(v,-1.0,1.0,v)
    # Now this generates on [0,pi)
    dec = numpy.arccos(v)

    # convert to degrees
    rad2deg(dec,dec)
    # now in range [-90,90.0)
    dec -= 90.0
    
    if system == 'xyz':
        x,y,z = eq2xyz(ra, dec)
        return x,y,z
    else:
        return ra, dec

def randcap(nrand, ra, dec, rad, get_radius=False):
    """
    Generate random points in a sherical cap

    parameters
    ----------

    nrand:
        The number of random points
    ra,dec:
        The center of the cap in degrees.  The ra should be within [0,360) and
        dec from [-90,90]
    rad:
        radius of the cap, same units as ra,dec

    get_radius: bool, optional
        if true, return radius of each point in radians
    """
    # generate uniformly in r**2
    rand_r = numpy.random.random(nrand)
    rand_r = sqrt(rand_r)*rad

    # put in degrees
    numpy.deg2rad(rand_r,rand_r)

    # generate position angle uniformly 0,2*PI
    rand_posangle = numpy.random.random(nrand)*2*PI

    theta = numpy.array(dec, dtype='f8',ndmin=1,copy=True)
    phi = numpy.array(ra,dtype='f8',ndmin=1,copy=True)
    theta += 90

    numpy.deg2rad(theta,theta)
    numpy.deg2rad(phi,phi)

    sintheta = sin(theta)
    costheta = cos(theta)
    sinphi = sin(phi)
    cosphi = cos(phi)

    sinr = sin(rand_r)
    cosr = cos(rand_r)

    cospsi = cos(rand_posangle)
    costheta2 = costheta*cosr + sintheta*sinr*cospsi

    numpy.clip(costheta2, -1, 1, costheta2)                    

    # gives [0,pi)
    theta2 = arccos(costheta2)
    sintheta2 = sin(theta2)

    cosDphi = (cosr - costheta*costheta2)/(sintheta*sintheta2)

    numpy.clip(cosDphi, -1, 1, cosDphi)                    
    Dphi = arccos(cosDphi)

    # note fancy usage of where
    phi2=numpy.where(rand_posangle > PI, phi+Dphi, phi-Dphi)

    numpy.rad2deg(phi2,phi2)
    numpy.rad2deg(theta2,theta2)
    rand_ra  = phi2
    rand_dec = theta2-90.0

    if get_radius:
        return rand_ra, rand_dec, rand_r
    else:
        return rand_ra, rand_dec


def rect_area(lon_min, lon_max, lat_min, lat_max):
    """
    Calculate the area of a rectangle on the sphere.

    parameters
    ----------
    lon_min, lon_max, lat_min, lat_max:
        Definition of the rectangle, in degrees
    """
    smax = sin(deg2rad(lat_max))
    smin = sin(deg2rad(lat_min))
    area = (smax-smin)*(lon_max-lon_min)
    return numpy.abs(area)*R2D



from __future__ import print_function, division
import numpy as np
from PyAstronomy.pyasl import idlMod
from PyAstronomy.pyaC import pyaErrors as PE
from PyAstronomy.pyasl import _ic


def sunpos2(jd, end_jd=None, jd_steps=None, outfile=None, radian=False, plot=False, full_output=False):
  """
    Compute right ascension and declination of the Sun at a given time.
    
    Parameters
    ----------
    jd : float
         The Julian date
    end_jd : float, optional
         The end of the time period as Julian date. If given,
         `sunpos` computes RA and DEC at `jd_steps` time points
         between `jd` and ending at `end_jd`.
    jd_steps : integer, optional
         The number of steps between `jd` and `end_jd`
         for which RA and DEC are to be calculated.
    outfile : string, optional
         If given, the output will be written to a file named according
         to `outfile`.
    radian : boolean, optional
         Results are returned in radian instead of in degrees.
         Default is False.
    plot : boolean, optional
         If True, the result is plotted.
    full_output: boolean, optional
         If True, `sunpos`, additionally, returns the elongation and
         obliquity of the Sun.

    Returns
    -------
    Time : array
        The JDs for which calculations where carried out.
    Ra : array
        Right ascension of the Sun.
    Dec : array
        Declination of the Sun.
    Elongation : array, optional
        Elongation of the Sun (only of `full_output`
        is set to True).
    Obliquity : array, optional
        Obliquity of the Sun (only of `full_output`
        is set to True).

    Notes
    -----
    
    .. note:: This function was ported from the IDL Astronomy User's Library.

    :IDL - Documentation:

    NAME:
          SUNPOS
    PURPOSE:
          To compute the RA and Dec of the Sun at a given date.
    
    CALLING SEQUENCE:
          SUNPOS, jd, ra, dec, [elong, obliquity, /RADIAN ]
    INPUTS:
          jd    - The Julian date of the day (and time), scalar or vector
                  usually double precision
    OUTPUTS:
          ra    - The right ascension of the sun at that date in DEGREES
                  double precision, same number of elements as jd
          dec   - The declination of the sun at that date in DEGREES
    
    OPTIONAL OUTPUTS:
          elong - Ecliptic longitude of the sun at that date in DEGREES.
          obliquity - the obliquity of the ecliptic, in DEGREES
    
    OPTIONAL INPUT KEYWORD:
          /RADIAN - If this keyword is set and non-zero, then all output variables 
                  are given in Radians rather than Degrees
    
    NOTES:
          Patrick Wallace (Rutherford Appleton Laboratory, UK) has tested the
          accuracy of a C adaptation of the sunpos.pro code and found the 
          following results.   From 1900-2100 SUNPOS  gave 7.3 arcsec maximum 
          error, 2.6 arcsec RMS.  Over the shorter interval 1950-2050 the figures
          were 6.4 arcsec max, 2.2 arcsec RMS.  
    
          The returned RA and Dec are in the given date's equinox.
    
          Procedure was extensively revised in May 1996, and the new calling
          sequence is incompatible with the old one.
    METHOD:
          Uses a truncated version of Newcomb's Sun.    Adapted from the IDL
          routine SUN_POS by CD Pike, which was adapted from a FORTRAN routine
          by B. Emerson (RGO).
    EXAMPLE:
          (1) Find the apparent RA and Dec of the Sun on May 1, 1982
          
          IDL> jdcnv, 1982, 5, 1,0 ,jd      ;Find Julian date jd = 2445090.5   
          IDL> sunpos, jd, ra, dec
          IDL> print,adstring(ra,dec,2)
                   02 31 32.61  +14 54 34.9
    
          The Astronomical Almanac gives 02 31 32.58 +14 54 34.9 so the error
                  in SUNPOS for this case is < 0.5".      
    
          (2) Find the apparent RA and Dec of the Sun for every day in 1997
    
          IDL> jdcnv, 1997,1,1,0, jd                ;Julian date on Jan 1, 1997
          IDL> sunpos, jd+ dindgen(365), ra, dec    ;RA and Dec for each day 
    
    MODIFICATION HISTORY:
          Written by Michael R. Greason, STX, 28 October 1988.
          Accept vector arguments, W. Landsman     April,1989
          Eliminated negative right ascensions.  MRG, Hughes STX, 6 May 1992.
          Rewritten using the 1993 Almanac.  Keywords added.  MRG, HSTX, 
                  10 February 1994.
          Major rewrite, improved accuracy, always return values in degrees
          W. Landsman  May, 1996 
          Added /RADIAN keyword,    W. Landsman       August, 1997
          Converted to IDL V5.0   W. Landsman   September 1997
  """
    
  if end_jd is None:
    # Form time in Julian centuries from 1900.0
    start_jd = (jd - 2415020.0)/36525.0
    # Zime array
    time = np.array(start_jd)
  else:
    if jd >= end_jd:
      raise(PE.PyAValError("`end_jd` needs to be larger than `jd`.", \
            where="sunpos", \
            solution="Modify the parameters."))
      if jd_steps is None:
        raise(PE.PyAValError("You specified `end_jd`, but no value for `jd_steps`.", \
              where="sunpos", \
              solution="Specify `jd_steps`, e.g., given jd_steps=10"))
    # Form time in Julian centuries from 1900.0
    start_jd = (jd - 2415020.0)/36525.0
    end_jd = (end_jd - 2415020.0)/36525.0    
    # Time array
    timestep = (end_jd-start_jd)/float(jd_steps)
    time = np.arange(start_jd, end_jd, timestep)

  # Mean solar longitude
  sunlon = (279.696678 + idlMod( (36000.768925*time), 360.0) )*3600.0

  # Allow for ellipticity of the orbit (equation of center)
  # using the Earth's mean anomaly ME
  me = 358.475844 + idlMod( (35999.049750*time) , 360.0 )
  ellcor  = ( 6910.1 - 17.2*time ) * np.sin(me*np.pi/180.) + 72.3 * np.sin(2.0*me*np.pi/180.)
  sunlon += ellcor

  # Allow for the Venus perturbations using the mean anomaly of Venus MV
  mv = 212.603219 + idlMod( (58517.803875*time) , 360.0 )
  vencorr = 4.8 * np.cos( (299.1017 + mv - me)*np.pi/180. ) + \
            5.5 * np.cos( (148.3133 +  2.0 * mv  -  2.0 * me )*np.pi/180. ) + \
            2.5 * np.cos( (315.9433 +  2.0 * mv  -  3.0 * me )*np.pi/180. ) + \
            1.6 * np.cos( (345.2533 +  3.0 * mv  -  4.0 * me )*np.pi/180. ) + \
            1.0 * np.cos( (318.15   +  3.0 * mv  -  5.0 * me )*np.pi/180. )
  sunlon += vencorr

  # Allow for the Mars perturbations using the mean anomaly of Mars MM
  mm = 319.529425  + idlMod( (19139.858500*time)  , 360.0 )
  marscorr = 2.0 * np.cos( (343.8883 -  2.0 * mm  +  2.0 * me)*np.pi/180. ) + \
            1.8 * np.cos( (200.4017 -  2.0 * mm  + me)*np.pi/180. )
  sunlon += marscorr

  # Allow for the Jupiter perturbations using the mean anomaly of Jupiter MJ
  mj = 225.328328  + idlMod( (3034.6920239*time) , 360.0 )
  jupcorr = 7.2 * np.cos( (179.5317 - mj + me )*np.pi/180. ) + \
            2.6 * np.cos( (263.2167 -  mj )*np.pi/180. ) + \
            2.7 * np.cos( ( 87.1450 -  2.0 * mj  +  2.0 * me )*np.pi/180. ) + \
            1.6 * np.cos( (109.4933 -  2.0 * mj  +  me )*np.pi/180. )
  sunlon += jupcorr

  # Allow for the Moon's perturbations using the mean elongation of
  # the Moon from the Sun D
  d = 350.7376814 + idlMod( (445267.11422*time) , 360.0 )
  mooncorr  = 6.5*np.sin(d*np.pi/180.)
  sunlon += mooncorr

  # Allow for long period terms
  longterm  = 6.4*np.sin( (231.19 + 20.20*time)*np.pi/180. )
  sunlon += longterm
  sunlon = idlMod( ( sunlon + 2592000.0 ) , 1296000.0 )
  longmed = sunlon/3600.0

  # Allow for Aberration
  sunlon -=  20.5

  # Allow for Nutation using the longitude of the Moons mean node OMEGA
  omega = 259.183275 - idlMod( (1934.142008*time) , 360.0 )
  sunlon = sunlon - 17.2*np.sin(omega*np.pi/180.)

  # Calculate the True Obliquity
  oblt = 23.452294 - 0.0130125*time + ( 9.2*np.cos(omega*np.pi/180.) )/3600.0

  # Right Ascension and Declination
  sunlon /= 3600.0
  ra = np.arctan2( np.sin(sunlon*np.pi/180.) * np.cos(oblt*np.pi/180.), np.cos(sunlon*np.pi/180.) )

  neg = np.where(ra < 0.0)[0]
  nneg = len(neg)
  if nneg > 0: ra[neg] += 2.0*np.pi

  dec = np.arcsin( np.sin(sunlon*np.pi/180.) * np.sin(oblt*np.pi/180.) )

  if radian:
    oblt *= (np.pi/180.)
    longmed *= (np.pi/180.)
  else:
    ra /= (np.pi/180.)
    dec /= (np.pi/180.)

  jd = time*36525.0 + 2415020.0

  if outfile is not None:
    # Write results to a file
    of = open(outfile, 'w')
    of.write("# File created by 'sunpos'\n")
    if not full_output:
      of.write("# 1) JD, 2) ra, 3) dec\n")
      np.savetxt(of, np.transpose(np.vstack((jd, ra, dec))))
    else:
      of.write("# 1) JD, 2) ra, 3) dec, 4) longitude, 5) obliquity\n")
      np.savetxt(of, np.transpose(np.vstack((jd, ra, dec, longmed, oblt))))
    of.close()

  if plot:
    if not _ic.check["matplotlib"]:
      raise(PE.PyARequiredImport("Could not import matplotlib.", \
            where="sunpos", \
            solution=["Install matplotlib", "Switch `plot` flag to False."]))
    import matplotlib.pylab as plt
    plt.plot(jd, ra, 'k-', label="RA")
    plt.plot(jd, dec, 'g-', label="DEC")
    plt.legend()
    plt.show()

  if full_output:
    return jd, ra, dec, longmed, oblt
  else:
    return jd, ra, dec



# encoding: utf-8

from __future__ import absolute_import, division, print_function, unicode_literals

import os

import yaml

# Inits the logging system. Only shell logging, and exception and warning catching.
# File logging can be started by calling log.start_file_logger(name).
# from .utils import get_logger


def merge(user, default):
    """Merges a user configuration with the default one."""

    if isinstance(user, dict) and isinstance(default, dict):
        for kk, vv in default.items():
            if kk not in user:
                user[kk] = vv
            else:
                user[kk] = merge(user[kk], vv)

    return user


NAME = 'roboscheduler'


# Loads config
config_file = os.path.dirname(__file__) + '/etc/{0}.yml'.format(NAME)
try:
    config = yaml.load(open(config_file), Loader=yaml.FullLoader)
except AttributeError:
    config = yaml.load(open(config_file))

# If there is a custom configuration file, updates the defaults using it.
custom_config_fn = os.path.expanduser('~/.{0}/{0}.yml'.format(NAME))
if os.path.exists(custom_config_fn):
    try:
        config = merge(yaml.load(open(custom_config_fn),
                                 Loader=yaml.FullLoader), config)
    except AttributeError:
        config = merge(yaml.load(open(custom_config_fn)), config)


__version__ = '0.10.0'



import numpy as np

"""Observations module class.

Dependencies:

 numpy
 scipy

"""


def wrapHA(ha):
    """Force ha into range -180 to 180. Just to be sure.
    """
    if -180 < ha <= 180:
        return ha
    while ha > 180:
        ha = ha - 360.
    while ha <= -180:
        ha = ha + 360.
    assert -180 < ha <= 180, "ha = {:.2f}".format(ha)
    return ha


class Observations(object):
    """Observations class.

    Parameters:
    ----------

    observatory : string
       Observatory for fields ('apo' or 'lco'; default 'apo')

    Attributes:
    ----------

    nobservations : np.int32
        number of observations

    fieldid : ndarray of np.int32
        id of each field for observations

    mjd : ndarray of np.float64
        MJD of observation (days)

    duration : ndarray of np.float64
        duration of observation (days)

    sn2 : ndarray of np.float32
        duration of observation (days)

    airmass : ndarray of np.float32
        airmass of observation

    skybrightness : ndarray of np.float32
        sky brightness of observation (Moon illumination fraction, or zero if
        below horizon, or one if in bright twilight)

    lst : ndarray of np.float32
        LST of observation

    Methods:
    -------

    add() : add an observation of a field
    toarray() : Return ndarray of field properties

"""
    def __init__(self, observatory='apo'):
        self.nobservations = np.int32(0)
        self.observatory = observatory
        self.fieldid = np.zeros(0, dtype=np.int32)
        self.racen = np.zeros(0, dtype=np.float64)
        self.deccen = np.zeros(0, dtype=np.float64)
        self.cadence = np.zeros(0, dtype=np.dtype('a20'))
        self.nfilled = np.zeros(0, dtype=np.int32)
        self.mjd = np.zeros(0, dtype=np.float64)
        self.duration = np.zeros(0, dtype=np.float64)
        self.apgSN2 = np.zeros(0, dtype=np.float32)
        self.rSN2 = np.zeros(0, dtype=np.float32)
        self.bSN2 = np.zeros(0, dtype=np.float32)
        self.airmass = np.zeros(0, dtype=np.float32)
        self.skybrightness = np.zeros(0, dtype=np.float32)
        self.lst = np.zeros(0, dtype=np.float32)
        self.ha = np.zeros(0, dtype=np.float64)
        self.nexp_cumul = np.zeros(0, dtype=np.int32)
        return

    def add(self, fieldid=None, mjd=None, duration=None, apgSN2=None,
            rSN2=None, bSN2=None, skybrightness=None, airmass=None,
            lst=None, racen=None, deccen=None, cadence=None, nfilled=None,
            nexp_cumul=None):
        self.fieldid = np.append(self.fieldid,
                                 np.array([np.float64(fieldid)]))
        self.mjd = np.append(self.mjd,
                             np.array([np.float64(mjd)]))
        self.duration = np.append(self.duration,
                                  np.array([np.float64(duration)]))
        self.apgSN2 = np.append(self.apgSN2,
                                np.array([np.float32(apgSN2)]))
        self.rSN2 = np.append(self.rSN2,
                              np.array([np.float32(rSN2)]))
        self.bSN2 = np.append(self.bSN2,
                              np.array([np.float32(bSN2)]))
        self.skybrightness = np.append(self.skybrightness,
                                       np.array([np.float32(skybrightness)]))
        self.airmass = np.append(self.airmass,
                                 np.array([np.float32(airmass)]))
        self.lst = np.append(self.lst,
                             np.array([np.float32(lst)]))
        self.nobservations = len(self.fieldid)

        self.racen = np.append(self.racen,
                               np.array([np.float64(racen)]))
        self.deccen = np.append(self.deccen,
                                np.array([np.float64(deccen)]))
        self.cadence = np.append(self.cadence,
                                 np.array([cadence]))
        self.nfilled = np.append(self.nfilled,
                                 np.array([np.int32(nfilled)]))

        ha = wrapHA(lst - racen)
        self.ha = np.append(self.ha,
                            np.array([np.float64(ha)]))
        self.nexp_cumul = np.append(self.nexp_cumul,
                                    np.array([np.int32(nexp_cumul)]))

        return(self.nobservations - 1)

    def forfield(self, mjd=None, fieldid=None):
        indx = np.where((self.mjd <= mjd) &
                        (self.fieldid == fieldid))[0]
        return(self.toarray(indx=indx))

    def toarray(self, indx=None):
        """Return observations as a record array

        Parameters:
        ----------

        indx : ndarray of np.int32
            indices of observations to return (default to all)

        Returns:
        -------

        observations : record array
            observation information
        """
        obs0 = [('fieldid', np.int32),
                ('mjd', np.float64),
                ('duration', np.float64),
                ('apgSN2', np.float32),
                ('rSN2', np.float32),
                ('bSN2', np.float32),
                ('airmass', np.float32),
                ('skybrightness', np.float32),
                ('lst', np.float32),
                ('racen', np.float64),
                ('deccen', np.float64),
                ('cadence', np.dtype('a20')),
                ('nfilled', np.int32),
                ('ha', np.float32),
                ('nexp_cumul', np.int32)]
        if(indx is None):
            indx = np.arange(self.nobservations)
        nobs = len(indx)
        obs = np.zeros(nobs, dtype=obs0)
        if(nobs > 0):
            obs['fieldid'] = self.fieldid[indx]
            obs['mjd'] = self.mjd[indx]
            obs['duration'] = self.duration[indx]
            obs['apgSN2'] = self.apgSN2[indx]
            obs['rSN2'] = self.rSN2[indx]
            obs['bSN2'] = self.bSN2[indx]
            obs['airmass'] = self.airmass[indx]
            obs['skybrightness'] = self.skybrightness[indx]
            obs['lst'] = self.lst[indx]
            obs['racen'] = self.racen[indx]
            obs['deccen'] = self.deccen[indx]
            obs['cadence'] = self.cadence[indx]
            obs['nfilled'] = self.nfilled[indx]
            obs['ha'] = self.ha[indx]
            obs['nexp_cumul'] = self.nexp_cumul[indx]
        return(obs)



import configparser
import pickle
import numpy as np
import fitsio
import roboscheduler.cCadenceCore as cCadenceCore

try:
    import sdssdb.peewee.sdss5db.targetdb as targetdb
    from sdssdb.peewee.sdss5db import database
    # database.set_profile('operations')
    _database = True
except:
    _database = False


def basename(cadence):
    return("_".join(cadence.split('-')[0].split('_')[0:-1]))


# Class to define a singleton
class CadenceListSingleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(CadenceListSingleton,
                                        cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class Cadence(cCadenceCore.CadenceCore):
    """Description of a cadence

    Parameters:
    ----------

    name : str
        name of cadence

    cfg : configparser.ConfigParser
        configuration object (if set all other parameters except 'name' ignored)

    nepochs : np.int32
        number of epochs

    skybrightness : ndarray of np.float32
        maximum sky brightness for each exposure

    delta : ndarray of np.float32
        desired offset for each exposure from previous (days)

    delta_min : ndarray of np.float32
        minimum delta to allow (days)

    delta_max : ndarray of np.float32
        maximum delta to allow (days)

    nexp : ndarray of np.int32
        number of exposures each epoch

    max_length : ndarray of np.float32
        max length from start of 1st exp to start of last (days)

    Attributes:
    ----------

    nexp_total : np.int32
        total number of exposures

    skybrightness : ndarray of np.float32
        maximum sky brightness for each exposure

    delta : ndarray of np.float32
        desired offset for each exposure from previous (days)

    delta_min : ndarray of np.float32
        minimum delta to allow (days)

    delta_max : ndarray of np.float32
        maximum delta to allow (days)

    nepochs : np.int32
        number of epochs

    epoch_indx : ndarray of np.int32
        index into delta for first exposure at each epoch

    epoch_nexposures : ndarray of np.int32
        number of exposures for each separate epoch

    epochs : ndarray of np.int32
        epoch number for each exposure

    Methods:
    -------

    Notes:
    -----

    The first delta == 0 in all cadences.

    delta == 0 for any exposure except the first means to "bundle" the
    exposure with the previous exposure into a single epoch. delta_min
    and delta_max for such exposures are automatically set to zero on
    input.

    The last epochs in the list may have delta = -1. These will be
    treated as unconstrained. Only single epoch cadences will be allowed.
    """
    def __init__(self, name=None, nepochs=None, skybrightness=None,
                 delta=None, delta_min=None, delta_max=None, nexp=None,
                 max_length=None, obsmode_pk=None, label_root=None,
                 label_version=None, cfg=None, min_moon_sep=None,
                 min_deltav_ks91=None, min_twilight_ang=None,
                 max_airmass=None, priorities={}):
        if(cfg is not None):
            self._from_cfg(name=name, cfg=cfg)
            return
        nepochs = np.int32(nepochs)
        skybrightness = self._arrayify(skybrightness, dtype=np.float32)
        delta = self._arrayify(delta, dtype=np.float32)
        delta_min = self._arrayify(delta_min, dtype=np.float32)
        delta_max = self._arrayify(delta_max, dtype=np.float32)
        nexp = self._arrayify(nexp, dtype=np.int32)
        max_length = self._arrayify(max_length, dtype=np.float32)
        min_moon_sep = self._arrayify(min_moon_sep, dtype=np.float32)
        min_deltav_ks91 = self._arrayify(min_deltav_ks91,
                                         dtype=np.float32)
        min_twilight_ang = self._arrayify(min_twilight_ang,
                                          dtype=np.float32)
        max_airmass = self._arrayify(max_airmass, dtype=np.float32)
        self.obsmode_pk = np.array(obsmode_pk, dtype=np.str)
        self.label_root = np.str(label_root)
        self.label_version = np.str(label_version)
        epoch_indx = np.zeros(nepochs + 1, dtype=np.int32)
        epochs = np.zeros(nexp.sum(), dtype=np.int32)
        super().__init__(name, nepochs,
                         skybrightness, delta, delta_min,
                         delta_max, nexp, max_length,
                         min_moon_sep, min_deltav_ks91,
                         min_twilight_ang, max_airmass,
                         epoch_indx, epochs)

        self.partialEpochPri = priorities.get("partialEpochPri", 200)
        self.brightDurDarkPenalty = priorities.get("brightDurDarkPenalty", -5000)
        self.deltaMaxPriBump = priorities.get("deltaMaxPriBump", 10)
        self.deltaNomBump = priorities.get("deltaNomBump", 4)
        self.overDeltaMaxBump = priorities.get("overDeltaMaxBump", 100)

        return

    def _from_cfg(self, name=None, cfg=None):
        nepochs = np.int32(cfg.get(name, 'nepochs'))
        skybrightness = np.array(cfg.get(name, 'skybrightness').split(),
                                 dtype=np.float32)
        delta = np.array(cfg.get(name, 'delta').split(),
                         dtype=np.float32)
        delta_min = np.array(cfg.get(name, 'delta_min').split(),
                             dtype=np.float32)
        delta_max = np.array(cfg.get(name, 'delta_max').split(),
                             dtype=np.float32)
        nexp = np.array(cfg.get(name, 'nexp').split(),
                        dtype=np.int32)
        max_length = np.array(cfg.get(name, 'max_length').split(),
                              dtype=np.float32)
        obsmode_pk = np.array(cfg.get(name, 'obsmode_pk').split(),
                              dtype=np.str)
        min_moon_sep = np.array(cfg.get(name, 'min_moon_sep').split(),
                                dtype=np.float32)
        min_deltav_ks91 = np.array(cfg.get(name, 'min_deltav_ks91').split(),
                                   dtype=np.float32)
        min_twilight_ang = np.array(cfg.get(name, 'min_twilight_ang').split(),
                                    dtype=np.float32)
        max_airmass = np.array(cfg.get(name, 'max_airmass').split(),
                               dtype=np.float32)
        label_root = np.str(cfg.get(name, 'label_root'))
        label_version = np.str(cfg.get(name, 'label_version'))

        self.__init__(name=name, nepochs=nepochs, skybrightness=skybrightness,
                      delta=delta, delta_min=delta_min, delta_max=delta_max,
                      nexp=nexp, max_length=max_length, obsmode_pk=obsmode_pk,
                      min_moon_sep=min_moon_sep, min_deltav_ks91=min_deltav_ks91,
                      min_twilight_ang=min_twilight_ang, max_airmass=max_airmass,
                      label_root=label_root, label_version=label_version)

        return

    def as_cadencecore(self):
        cc = cCadenceCore.CadenceCore(self.name, self.nepochs,
                                      self.skybrightness, self.delta,
                                      self.delta_min, self.delta_max, self.nexp,
                                      self.max_length,
                                      self.min_moon_sep,
                                      self.min_deltav_ks91,
                                      self.min_twilight_ang,
                                      self.max_airmass,
                                      self.epoch_indx, self.epochs)
        return(cc)

    def _arrayify(self, quantity=None, dtype=np.float64):
        """Cast quantity as ndarray of numpy.float64"""
        try:
            length = len(quantity)
        except TypeError:
            length = 1
        return np.zeros(length, dtype=dtype) + quantity

    def skybrightness_check(self, epoch_idx, skybrightness_next):
        """check lunation for mjd_past against lunation_next"""
        down_weight = False
        if(epoch_idx >= self.nepochs):
            if self.skybrightness[-1] > 0.4:
                if skybrightness_next < 0.4:
                    down_weight = True
            return skybrightness_next <= self.skybrightness[-1], down_weight
        if self.skybrightness[epoch_idx] > 0.4:
            if skybrightness_next < 0.4:
                down_weight = True
        return skybrightness_next <= self.skybrightness[epoch_idx], down_weight

    def obsmodeChecks(self, epoch_idx, moon_dist, deltaV, airmass):
        """check obsmode reqs: moon sep, deltav, & airmass
        """
        if epoch_idx >= self.nepochs:
            epoch_idx = -1
        return self.min_deltav_ks91[epoch_idx] <= deltaV and \
            self.max_airmass[epoch_idx] >= airmass and \
            self.min_moon_sep[epoch_idx] <= moon_dist

    def evaluate_next(self, epoch_idx=None, partial_epoch=False,
                      mjd_past=None, mjd_next=None,
                      skybrightness_next=None, moon_dist=None, deltaV=None,
                      airmass=None):
        """Evaluate next choice of observation

           Returns whether cadence is ok AND how long until
           cadence will become impossible (deltaMax - delta)
        """

        if partial_epoch:
            base_priority = self.partialEpochPri
        elif(epoch_idx >= self.nepochs):
            print("done!")
            return(False, 0)
        else:
            base_priority = 0

        ok_skybrightness, down_weight = self.skybrightness_check(epoch_idx,
                                                                 skybrightness_next)

        valid = self.obsmodeChecks(epoch_idx, moon_dist, deltaV, airmass)
        ok_skybrightness = ok_skybrightness & valid

        if not ok_skybrightness:
            return ok_skybrightness, 0

        if down_weight:
            base_priority = self.brightDurDarkPenalty

        if(epoch_idx == 0):
            # if "x8" in self.name:
            #     print(self.name, "FIRST!", ok_skybrightness)
            return(ok_skybrightness, base_priority)

        delta_curr = mjd_next - mjd_past
        dlo = self.delta_min[epoch_idx]
        dhi = self.delta_max[epoch_idx]
        dnom = self.delta[epoch_idx]
        if(dlo == -1):
            # if "x8" in self.name:
            #     print(self.name, "-1?", ok_skybrightness)
            return(ok_skybrightness, base_priority)
        # if "x1" in self.name:
        #     print("delta {} dhi {} dlo {} curr {:.2f}".format(dnom, dhi, dlo, float(delta_curr)), self.name)
        # 1/sqrt(x) priority; at 1 day +100, at 10 days +30, at 30 days +18
        remain_priority = self.deltaMaxPriBump * np.clip(10/np.sqrt(np.abs(dhi - delta_curr)),
                                       a_min=None, a_max=10)
        nom_priority = self.deltaNomBump * np.clip(10/np.sqrt(np.abs(dnom - delta_curr)),
                                   a_min=None, a_max=10)
        priority = base_priority + remain_priority + nom_priority
        if delta_curr <= dhi:
            priority += self.overDeltaMaxBump
        # if ignoreMax:
        #     return(ok_skybrightness & (delta_curr >= dlo), priority)
        # if "x8" in self.name:
        #     print("CAD", self.name, ok_skybrightness, delta_curr, dlo, priority)
        return(ok_skybrightness & (delta_curr >= dlo), priority)


class CadenceList(object, metaclass=CadenceListSingleton):
    """List of cadences available (singleton)

    Parameters:
    ----------

    skybrightness_only : bool
        Only use sky brightness for consistency checks (default False)

    Attributes:
    ----------

    skybrightness_only : bool
        Only use sky brightness for consistency checks

    ncadences : np.int32, int
        number of different cadences

    cadences : dict
        dictionary of Cadence objects

    Methods:
    -------

    reset() : remove all the current cadences
    add_cadence() : add a new cadence
    fromarray(): add to cadence list from an ndarray
    fromfits(): add to cadence list from a FITS file
    fromcfg(): add to cadence list from a configuration file
    toarray(): return an ndarray with cadence list
    epoch_array(): return an ndarray with epoch-oriented list of cadences
    todb(): insert cadences into the targetdb
    fromdb(): extract cadences into the targetdb

    Notes:
    -----

    This is a singleton, so there can only be one CadenceList defined
    within any session.
    """
    def __init__(self, skybrightness_only=False):
        self.reset()
        self.max_nsolns = 100
        self.skybrightness_only = skybrightness_only
        return

    def reset(self):
        """Reset cadence list to be empty"""
        self.ncadences = 0
        self.cadences = dict()
        self._cadence_consistency = dict()
        return

    def add_cadence(self, name=None, *args, **kwargs):
        """Add a cadence to the list of cadences

        Parameters:
        ----------

        name : string
            dictionary name of cadence

        nepochs : np.int32
            number of epochs (default 1)

        skybrightness : ndarray of np.float32
            maximum sky brightness for each exposure (default [1.])

        delta : ndarray of np.float32
            day for exposure (default [0.])

        delta_min : ndarray of np.float32
            allowance for variation from cadence (default [1.])

        delta_max : ndarray of np.float32
            allowance for variation from cadence (default [1.])

        nexp : ndarray of np.int32
            number of exposures per cadence

        max_length : ndarray of np.float32
            max length from start of 1st exp to start of last (days)

        Notes:
        -----

        The last epochs in the list may have delta = -1. These will be
        treated as unconstrained. Only single epoch cadences will be allowed.
        """
        cadence = Cadence(name=name, *args, **kwargs)
        self.cadences[name] = cadence
        self.ncadences = len(self.cadences.keys())
        return

    def cadence_consistency(self, one, two, return_solutions=True,
                            epoch_level=True, merge_epochs=False):
        """Is cadence #1 consistent with cadence #2?

        Parameters:
        ----------

        one : string
            name of cadence #1

        two : string
            name of cadence #2

        return_solutions: boolean
            return list of solutions? (default False)

        epoch_level : boolean
            compare sequences at epoch level not exposure level (default True)
            [ignores this]

        merge_epochs : bool
            if epochs in the solutions are repeated, merge them

        Returns:
        -------

        ok : int
            1 if there is a solution, 0 otherwise

        solutions : list of ndarrays of np.int32
            list of solutions, where each solution is an array

        nexps : list of ndarrays of np.int32
            if epochs are merged, number of exposures for each epoch listed

        Notes:
        -----

        When epochs are merged with merge_epochs you need to keep track
        of how many exposures you are putting into each epoch. In this 
        case 'nexps' is returned.
        """
        cache_key = (one, two, epoch_level, return_solutions,
                     merge_epochs, self.skybrightness_only)
        if(cache_key in self._cadence_consistency):
            return(self._cadence_consistency[cache_key])

        onecc = self.cadences[one].as_cadencecore()
        possibles = self.cadences[two].cadence_consistency(onecc,
                                                           self.skybrightness_only)
        success = len(possibles) > 0

        if(return_solutions):
            if(merge_epochs):
                possibles_merge = []
                nexps_merge = []
                for possible in possibles:
                    possible_merge, possible_inverse = np.unique(possible,
                                                                 return_inverse=True)
                    nexp_one = self.cadences[one].nexp
                    nexp_merge = np.zeros(len(possible_merge), dtype=np.int32)
                    for i, nexp in zip(possible_inverse, nexp_one):
                        nexp_merge[i] = nexp_merge[i] + nexp

                    possibles_merge.append(possible_merge)
                    nexps_merge.append(nexp_merge)


                self._cadence_consistency[cache_key] = (success,
                                                        possibles_merge,
                                                        nexps_merge)
            else:
                self._cadence_consistency[cache_key] = (success, possibles)
        else:
            self._cadence_consistency[cache_key] = success

        return(self._cadence_consistency[cache_key])

    def exposure_consistency(self, one, two, iexp):
        """Is cadence #1 consistent with a set of exposures in cadence #2?

        Parameters:
        ----------

        one : string
            name of cadence #1

        two : string
            name of cadence #2

        iexp : ndarray of np.int32
            indices of exposures to check

        Returns:
        -------

        ok : bool
            True if these exposures under cadence #2 satisfy cadence #1
        """
        # Are there any ways cadence one fits into cadence two?
        ok, epochs_list = self.cadence_consistency(one, two)
        if(not ok):
            return False

        cadence_two = self.cadences[two]
        cadence_one = self.cadences[one]

        # Now count how many exposures there are at each epoch for this
        # array of iexp
        epochs_two_got, nexps_got = np.unique(cadence_two.epochs[iexp],
                                              return_counts=True)
        ngot = np.zeros(cadence_two.nepochs, dtype=np.int32)
        ngot[epochs_two_got] = nexps_got

        # Check each solution and see if the current assignment satisfies

        nneed = np.zeros(cadence_two.nepochs, dtype=np.int32)
        for epochs_two in epochs_list:

            # For this solution, count how many exposures are required
            # at each epoch of cadence two
            nneed[:] = 0
            bad = False
            for epoch_one, epoch_two in enumerate(epochs_two):
                nneed[epoch_two] = nneed[epoch_two] + cadence_one.nexp[epoch_one]
                if(nneed[epoch_two] > ngot[epoch_two]):
                    bad = True
                    break

            # For each exposure in the solution, are there enough exposures?
            if(not bad):
                nbad = (ngot < nneed).sum()
                if(nbad == 0):
                    return True

        # If we have gotten here no working solution was found
        return False

    def fromarray(self, cadences_array=None, priorities={}):
        """Add cadences to ccadence list from an array

        Parameters:
        -----------

        cadences_array : ndarray
            ndarray with columns 'NEPOCHS', 'SKYBRIGHTNESS', 'DELTA',
            'DELTA_MIN', 'DELTA_MAX', 'NEXP', 'CADENCE',
            'MAX_LENGTH'

        """
        for ccadence in cadences_array:
            nepochs = ccadence['NEPOCHS']

            self.add_cadence(name=ccadence['CADENCE'],
                             nepochs=ccadence['NEPOCHS'],
                             skybrightness=ccadence['SKYBRIGHTNESS'][0:nepochs],
                             delta=ccadence['DELTA'][0:nepochs],
                             delta_min=ccadence['DELTA_MIN'][0:nepochs],
                             delta_max=ccadence['DELTA_MAX'][0:nepochs],
                             nexp=ccadence['NEXP'][0:nepochs],
                             max_length=ccadence['MAX_LENGTH'][0:nepochs],
                             obsmode_pk=ccadence['OBSMODE_PK'][0:nepochs],
                             min_moon_sep=ccadence['MIN_MOON_SEP'][0:nepochs],
                             min_deltav_ks91=ccadence['MIN_DELTAV_KS91'][0:nepochs],
                             min_twilight_ang=ccadence['MIN_TWILIGHT_ANG'][0:nepochs],
                             max_airmass=ccadence['MAX_AIRMASS'][0:nepochs],
                             label_root=ccadence['LABEL_ROOT'],
                             label_version=ccadence['LABEL_VERSION'],
                             priorities=priorities)
        return

    def fromfits(self, filename=None, unpickle=False, priorities={}):
        """Add cadences to ccadence list from a FITS file

        Parameters:
        -----------

        filename : str
            File name to read from

        Notes:
        -----

        Expects a valid FITS file with columns 'NEPOCHS',
            'SKYBRIGHTNESS', 'DELTA', 'DELTA_MIN', 'DELTA_MAX', 'CADENCE',
            'NEXP'
        """
        self.cadences_fits = fitsio.read(filename)
        self.fromarray(self.cadences_fits, priorities=priorities)
        if(unpickle):
            fp = open(filename + '.pkl', 'rb')
            cc = pickle.load(fp)
            for key in cc:
                self._cadence_consistency[key] = cc[key]
        return

    def fromcfg(self, filename=None):
        """Add cadences to cadence list from a configuration file

        Parameters:
        -----------

        filename : str
            File name to read from

        Notes:
        -----

        Expects a configuration file compatible with configparser. Each
        section corresponds to a cadence, and is specified as follows:

        [cadence_name]
        nepochs = <N>
        skybrightness = <sb1> .. <sbN>
        delta = <delta1> .. <deltaN>
        delta_min = <dmin1> .. <dminN>
        delta_max = <dmax1> .. <dminN>
        nexp = <nexp1> .. <nexpN>
        max_length = <ml1> .. <mlN>
        min_moon_sep = <ml1> .. <mlN>
        min_deltav_ks91 = <ml1> .. <mlN>
        min_twilight_ang = <ml1> .. <mlN>
        max_airmass = <ml1> .. <mlN>
        """

        cfg = configparser.ConfigParser()
        cfg.read(filename)
        for name in cfg.sections():
            self.add_cadence(name=name, cfg=cfg)
        return

    def toarray(self):
        """Return cadences as a record array

        Returns:
        -------

        cadences : ndarray
            information on each cadence
        """
        nepochs = np.array([c.nepochs for c in self.cadences.values()])
        max_nexp = nepochs.max()
        cadence0 = [('CADENCE', np.unicode_, 40),
                    ('NEPOCHS', np.int32),
                    ('DELTA', np.float64, max_nexp),
                    ('SKYBRIGHTNESS', np.float32, max_nexp),
                    ('DELTA_MAX', np.float32, max_nexp),
                    ('DELTA_MIN', np.float32, max_nexp),
                    ('NEXP', np.int32, max_nexp),
                    ('MAX_LENGTH', np.float32, max_nexp),
                    ('OBSMODE_PK', np.dtype("|S15"), max_nexp),
                    ('MIN_MOON_SEP', np.float32, max_nexp),
                    ('MIN_DELTAV_KS91', np.float32, max_nexp),
                    ('MIN_TWILIGHT_ANG', np.float32, max_nexp),
                    ('MAX_AIRMASS', np.float32, max_nexp),
                    ('LABEL_ROOT', np.unicode_, 40),
                    ('LABEL_VERSION', np.unicode_, 40)]
        cads = np.zeros(self.ncadences, dtype=cadence0)
        names = self.cadences.keys()
        for indx, name in enumerate(names):
            nepochs = self.cadences[name].nepochs
            cads['CADENCE'][indx] = name
            cads['NEPOCHS'][indx] = nepochs
            cads['DELTA'][indx, 0:nepochs] = self.cadences[name].delta[0:nepochs]
            cads['DELTA_MIN'][indx, 0:nepochs] = self.cadences[name].delta_min[0:nepochs]
            cads['DELTA_MAX'][indx, 0:nepochs] = self.cadences[name].delta_max[0:nepochs]
            cads['NEXP'][indx, 0:nepochs] = self.cadences[name].nexp[0:nepochs]
            cads['MAX_LENGTH'][indx, 0:nepochs] = self.cadences[name].max_length[0:nepochs]
            cads['OBSMODE_PK'][indx, 0:nepochs] = self.cadences[name].obsmode_pk
            cads['MIN_MOON_SEP'][indx, 0:nepochs] = self.cadences[name].min_moon_sep[0:nepochs]
            cads['MIN_DELTAV_KS91'][indx, 0:nepochs] = self.cadences[name].min_deltav_ks91[0:nepochs]
            cads['MIN_TWILIGHT_ANG'][indx, 0:nepochs] = self.cadences[name].min_twilight_ang[0:nepochs]
            cads['MAX_AIRMASS'][indx, 0:nepochs] = self.cadences[name].max_airmass[0:nepochs]
            cads['SKYBRIGHTNESS'][indx][0:nepochs] = self.cadences[name].skybrightness[0:nepochs]
            cads['LABEL_ROOT'][indx] = self.cadences[name].label_root
            cads['LABEL_VERSION'][indx] = self.cadences[name].label_version
        return(cads)

    def tocsv(self, filename=None):
        """write cadences to a csv file

        Parameters:
        -----------

        filename : str
            File name to write csv file to

        Returns:
        -------

        cadences : ndarray
            information on each cadence
        """
        names = self.cadences.keys()
        with open(filename, 'w', encoding='utf-8-sig', newline='') as csvfile:
            for indx, name in enumerate(names):
                nepochs = self.cadences[name].nepochs
                label_root = self.cadences[name].label_root
                label_version = self.cadences[name].label_version
                delta_string = np.array2string(self.cadences[name].delta[0:nepochs], precision=3, separator=',')
                delta_trim = delta_string.strip("[] ")
                delta_min_string = np.array2string(self.cadences[name].delta_min[0:nepochs], precision=3, separator=',')
                delta_min_trim = delta_min_string.strip("[] ")
                delta_max_string = np.array2string(self.cadences[name].delta_max[0:nepochs], precision=3, separator=',')
                delta_max_trim = delta_max_string.strip("[] ")
                nexp_string = np.array2string(self.cadences[name].nexp[0:nepochs], precision=3, separator=',')
                nexp_trim = nexp_string.strip("[] ")
                max_length_string = np.array2string(self.cadences[name].max_length[0:nepochs], precision=3, separator=',')
                max_length_trim = max_length_string.strip("[] ")
                x = self.cadences[name].obsmode_pk[0:nepochs]
                obsmode_pk_string = np.array2string(x, separator='","', formatter={'str_kind': lambda x: x})
                obsmode_pk_trim = '"'+obsmode_pk_string.strip("[] ")+'"'
                #print(obsmode_pk_trim)
                sky_brightness_string = np.array2string(self.cadences[name].skybrightness[0:nepochs], precision=3, separator=',')
                sky_brightness_trim = sky_brightness_string.strip("[] ")
                print_string = name + ';' + str(nepochs) + ';{' + delta_trim + '}' +\
                                      ';{' + sky_brightness_trim + '}' +\
                                      ';{' + delta_max_trim+'}' +\
                                      ';{' + delta_min_trim+'}' + ';{'+nexp_trim+'}' +\
                                      ';{' + max_length_trim+'}' + ';{'+obsmode_pk_trim+'};' +\
                                      label_root+';'+label_version
                clean_string = print_string.replace("\r","")
                cleaner_string = clean_string.replace("\n","")
                csvfile.write(cleaner_string+' \n')

    def todb(self):
        """Insert all cadences into the targetdb

        Notes:
        -----

        This may be obsolete. (John uses it locally and likes it)
        """

        if(_database is False):
            print("No database available.")
            return()

        obsModes = targetdb.ObsMode.select().dicts()

        DBmodes = {str(o["label"]): o for o in obsModes}

        for cadence in self.cadences:
            if targetdb.Cadence.get_or_none(label=cadence) is not None:
                continue
            nexposures = [int(n) for n in self.cadences[cadence].nexp]
            delta = [float(n) for n in self.cadences[cadence].delta]
            delta_min = [float(n) for n in self.cadences[cadence].delta_min]
            delta_max = [float(n) for n in self.cadences[cadence].delta_max]
            skybrightness = [float(n) for n in self.cadences[cadence].skybrightness]
            max_length = [float(n) for n in self.cadences[cadence].max_length]
            obs_modes = [str(n) for n in self.cadences[cadence].obsmode_pk]
            if all([s < 0.1 for s in skybrightness]):
                # not sure why this is needed, but seems to be with some fits files
                continue
            assert all([o in DBmodes for o in obs_modes]), \
                "obsmode_pks must be in obsmode table \n" +\
                "".join(np.unique(np.extract([o not in DBmodes for o in obs_modes], obs_modes)))\
                + "\n are not present"
            targetdb.Cadence.insert(label=cadence,
                                    nexp=nexposures,
                                    nepochs=self.cadences[cadence].nepochs,
                                    delta=delta, skybrightness=skybrightness,
                                    delta_min=delta_min,
                                    delta_max=delta_max,
                                    max_length=max_length,
                                    obsmode_pk=obs_modes).execute()

    def updatedb(self):
        """Update the cadences in the targetdb by name

        Notes:
        -----

        This may be obsolete.
        """
        if(_database is False):
            print("No database available.")
            return()

        for cadence in self.cadences:
            update_dict = {targetdb.Cadence.nexposures:
                           self.cadences[cadence].nexposures,
                           targetdb.Cadence.delta:
                           [float(n) for n in self.cadences[cadence].delta],
                           targetdb.Cadence.delta_min:
                           [float(n)
                            for n in self.cadences[cadence].delta_min],
                           targetdb.Cadence.delta_max:
                           [float(n)
                            for n in self.cadences[cadence].delta_max],
                           targetdb.Cadence.skybrightness:
                           [float(n) for n in self.cadences[cadence].skybrightness],
                           targetdb.Cadence.epoch_max_length:
                           [float(n)
                            for n in self.cadences[cadence].max_length]}
            targetdb.Cadence.update(update_dict). \
                where(targetdb.Cadence.label == cadence).execute()
        return

    def fromdb(self, use_label_root=True, version='', cadences=None):
        """Extract cadences into the targetdb

        Parameters:
        ----------

        cadences : list of str
            list of cadences to import from db (based on label_root);
            if None, imports all (default None)

        version : str
            restrict to these versions; if None, no restrictions
            (default '')

        use_label_root : bool
            use the label_root as the cadence name (default True)
"""
        if(_database is False):
            print("No database available.")
            return()

        cadence_dicts = targetdb.Cadence.select().dicts()
        obsmode_dicts = targetdb.ObsMode.select().dicts()
        obsmodes = dict()
        for obsmode_dict in obsmode_dicts:
            curr = dict()
            curr['min_deltav_ks91'] = obsmode_dict['min_deltav_ks91']
            curr['min_moon_sep'] = obsmode_dict['min_moon_sep']
            curr['min_twilight_ang'] = obsmode_dict['min_twilight_ang']
            curr['max_airmass'] = obsmode_dict['max_airmass']
            obsmodes[obsmode_dict['label']] = curr

        for cadence in cadence_dicts:
            if(cadence['delta'] is None or len(cadence['delta']) == 0):
                continue

            if(version is not None):
                if(version == ''):
                    if(cadence['label_version'] != ''):
                        continue
                else:
                    if(cadence['label_version'] != '_' + version):
                        continue

            label_root = str(cadence['label_root'].strip())
            if(cadences is not None):
                if(label_root not in cadences):
                    continue

            if(use_label_root):
                label = label_root
            else:
                label = str(cadence['label'])

            if(cadence['obsmode_pk'] is not None):
                min_moon_sep = np.array([obsmodes[x]['min_moon_sep']
                                         for x in cadence['obsmode_pk']],
                                        dtype=np.float32)
                min_deltav_ks91 = np.array([obsmodes[x]['min_deltav_ks91']
                                            for x in cadence['obsmode_pk']],
                                           dtype=np.float32)
                min_twilight_ang = np.array([obsmodes[x]['min_twilight_ang']
                                             for x in cadence['obsmode_pk']],
                                            dtype=np.float32)
                max_airmass = np.array([obsmodes[x]['max_airmass']
                                        for x in cadence['obsmode_pk']],
                                       dtype=np.float32)
            else:
                min_moon_sep = np.zeros(np.int32(cadence['nepochs']),
                                        dtype=np.float32)
                min_deltav_ks91 = np.zeros(np.int32(cadence['nepochs']),
                                           dtype=np.float32)
                min_twilight_ang = np.zeros(np.int32(cadence['nepochs']),
                                            dtype=np.float32)
                max_airmass = np.zeros(np.int32(cadence['nepochs']),
                                       dtype=np.float32)
                for epoch in np.arange(np.int32(cadence['nepochs']),
                                       dtype=np.int32):
                    if(cadence['skybrightness'][epoch] < 0.5):
                        min_moon_sep[epoch] = 15.
                        min_deltav_ks91[epoch] = - 2.5
                        min_twilight_ang[epoch] = 8.
                        max_airmass[epoch] = 2.
                    else:
                        min_moon_sep[epoch] = 35.
                        min_deltav_ks91[epoch] = - 1.5
                        min_twilight_ang[epoch] = 15.
                        max_airmass[epoch] = 1.4

            self.add_cadence(name=label,
                             nexp=cadence['nexp'],
                             nepochs=cadence['nepochs'],
                             delta=np.array(cadence['delta']),
                             delta_min=np.array(cadence['delta_min']),
                             delta_max=np.array(cadence['delta_max']),
                             skybrightness=np.array(cadence['skybrightness']),
                             max_length=np.array(cadence['max_length']),
                             min_moon_sep=min_moon_sep,
                             min_deltav_ks91=min_deltav_ks91,
                             min_twilight_ang=min_twilight_ang,
                             max_airmass=max_airmass,
                             obsmode_pk=np.array(cadence['obsmode_pk']),
                             label_root=str(cadence['label_root'].strip()),
                             label_version=str(cadence['label_root'].strip()))



# encoding: utf-8
#
# setup.py
#

from setuptools import setup, Extension, find_packages

import os
import argparse
import sys


class getPybindInclude(object):
    """Helper class to determine the pybind11 include path
    The purpose of this class is to postpone importing pybind11
    until it is actually installed, so that the ``get_include()``
    method can be invoked.
    https://github.com/pybind/python_example/blob/master/setup.py
    """

    def __init__(self, user=False):
        self.user = user

    def __str__(self):
        import pybind11
        return pybind11.get_include(self.user)


def getIncludes():
    return [
        'include',
        getPybindInclude(),

    ]


sources = [
    'cextern/cCadenceCore.cpp',
    'cextern/cadenceCore.cpp'
]

extra_compile_args = ["--std=c++11", "-fPIC", "-v", "-O3"]
extra_link_args = None
if sys.platform == 'darwin':
    extra_compile_args += ['-stdlib=libc++', '-mmacosx-version-min=10.9']
    extra_link_args = ["-v", '-mmacosx-version-min=10.9']

module = Extension(
    'roboscheduler.cCadenceCore',
    include_dirs=getIncludes(),
    extra_compile_args=extra_compile_args,
    extra_link_args=extra_link_args,
    sources=sources
)

setup(ext_modules=[module])
